######################################################.
#          This file stores functions used           #
#                in multiple modules                 #
######################################################.

import os
import subprocess
import sys
import time
import getopt
import numpy as np
import glob
import yaml
import ast
from pathlib import Path
from rdkit.Chem.rdMolAlign import GetBestRMS
from rdkit.Chem.rdmolops import RemoveHs
from rdkit.Chem import Mol
from rdkit.Chem import AllChem as Chem
from aqme.argument_parser import set_options, var_dict
from rdkit import RDLogger

GAS_CONSTANT = 8.3144621  # J / K / mol
J_TO_AU = 4.184 * 627.509541 * 1000.0  # UNIT CONVERSION
T = 298.15

aqme_version = "1.4.2"
time_run = time.strftime("%Y/%m/%d %H:%M:%S", time.localtime())
aqme_ref = f"AQME v {aqme_version}, Alegre-Requena, J. V.; Sowndarya, S.; Perez-Soto, R.; Alturaifi, T. M.; Paton, R. S., 2022. https://github.com/jvalegre/aqme"

RDLogger.DisableLog("rdApp.*")


def run_command(command, outfile):
    """
    Runs the subprocess command and outputs to the necessary output file
    """

    p2 = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
    txt = [line.decode("utf-8") for line in p2.stdout]
    p2.stdout.close()

    with open(outfile, "w") as f:
        for line in txt:
            f.write(line)
    f.close()


def periodic_table():
    items = """X
			H                                                                                                  He
			Li Be  B                                                                             C   N   O   F  Ne
			Na Mg Al                                                                            Si   P   S  Cl  Ar
			K Ca Sc                                           Ti  V Cr Mn Fe Co Ni Cu  Zn  Ga  Ge  As  Se  Br  Kr
			Rb Sr  Y                                           Zr Nb Mo Tc Ru Rh Pd Ag  Cd  In  Sn  Sb  Te   I  Xe
			Cs Ba La Ce Pr Nd Pm Sm Eu Gd Tb Dy Ho Er Tm Yb Lu Hf Ta  W Re Os Ir Pt Au  Hg  Tl  Pb  Bi  Po  At  Rn
			Fr Ra Ac Th Pa  U Np Pu Am Cm Bk Cf Es Fm Md No Lr Rf Db Sg Bh Hs Mt Ds Rg Uub Uut Uuq Uup Uuh Uus Uuo
			"""
    periodic_table = items.replace("\n", " ").strip().split()
    periodic_table[0] = ""

    return periodic_table


# load paramters from yaml file
def load_from_yaml(self):
    """
    Loads the parameters for the calculation from a yaml if specified. Otherwise
    does nothing.
    """

    txt_yaml = f"\no  Importing AQME parameters from {self.varfile}"
    error_yaml = False
    # Variables will be updated from YAML file
    try:
        if os.path.exists(self.varfile):
            if os.path.splitext(self.varfile)[1] in [".yaml", ".yml", ".txt"]:
                with open(self.varfile, "r") as file:
                    try:
                        param_list = yaml.load(file, Loader=yaml.SafeLoader)
                    except yaml.scanner.ScannerError:
                        txt_yaml = f'\nx  Error while reading {self.varfile}. Edit the yaml file and try again (i.e. use ":" instead of "=" to specify variables)'
                        error_yaml = True
        if not error_yaml:
            for param in param_list:
                if hasattr(self, param):
                    if getattr(self, param) != param_list[param]:
                        setattr(self, param, param_list[param])

    except UnboundLocalError:
        txt_yaml = "\nx  The specified yaml file containing parameters was not found! Make sure that the valid params file is in the folder where you are running the code."

    return self, txt_yaml


# class for logging
class Logger:
    """
    Class that wraps a file object to abstract the logging.
    """

    # Class Logger to writargs.input.split('.')[0] output to a file
    def __init__(self, filein, append, suffix="dat"):
        self.log = open(f"{filein}_{append}.{suffix}", "w")

    def write(self, message):
        """
        Appends a newline character to the message and writes it into the file.

        Parameters
        ----------
        message : str
           Text to be written in the log file.
        """
        self.log.write(f"{message}\n")
        print(f"{message}\n")

    def fatal(self, message):
        """
        Writes the message to the file. Closes the file and raises an error exit

        Parameters
        ----------
        message : str
           text to be written in the log file.
        """
        self.write(message)
        self.finalize()
        raise SystemExit(1)

    def finalize(self):
        """
        Closes the file
        """
        self.log.close()


def move_file(destination, source, file):
    """
    Moves files from the source folder to the destination folder and creates
    the destination folders when needed.

    Parameters
    ----------
    destination : str
        Path to the destination folder
    src : str
        Path to the source folder
    file : str
        Full name of the file (file + extension)
    """

    destination.mkdir(exist_ok=True, parents=True)
    filepath = source / file
    try:
        filepath.rename(destination / file)
    except FileExistsError:
        filepath.replace(destination / file)


def get_info_input(file):
    """
    Takes an input file and retrieves the coordinates of the atoms and the
    total charge.

    Parameters
    ----------
    file : str or pathlib.Path
        A path pointing to a valid .com or .gjf file

    Returns
    -------
    coordinates : list
        A list of strings (without \\n) that contain the xyz coordinates of the .gjf or .com file
    charge : str
        A str with the number corresponding to the total charge of the .com or .gjf file
    """

    with open(file, "r") as input_file:
        input_lines = input_file.readlines()

    _iter = input_lines.__iter__()

    line = ""
    # input for Gaussian calculations
    if str(file).split(".")[1] in ["com", "gjf"]:

        # Find the command line
        while "#" not in line:
            line = next(_iter)

        # in case the keywords are distributed in multiple lines
        while len(line.split()) > 0:
            line = next(_iter)

        # pass the title lines
        _ = next(_iter)
        while line:
            line = next(_iter).strip()

        # Read charge and multiplicity
        charge, mult = next(_iter).strip().split()

        # Store the atom types and coordinates until next empty line.
        atoms_and_coords = []
        line = next(_iter).strip()
        while line:
            atoms_and_coords.append(line.strip())
            line = next(_iter).strip()

    # input for ORCA calculations
    if str(file).split(".")[1] == ".inp":

        # Find the line with charge and multiplicity
        while "* xyz" not in line or "* int" not in line:
            line = next(_iter)

        # Read charge and multiplicity
        charge = line.strip().split()[-2]

        # Store the coordinates until next *
        atoms_and_coords = []
        line = next(_iter).strip()
        while len(line.split()) > 1:
            atoms_and_coords.append(line.strip())
            line = next(_iter).strip()
    return atoms_and_coords, charge, mult


def rules_get_charge(mol, args):
    """
    Automatically sets the charge for metal complexes
    """

    C_group = ["C", "Se", "Ge"]
    N_group = ["N", "P", "As"]
    O_group = ["O", "S", "Se"]
    F_group = ["F", "Cl", "Br", "I"]
    
    
    M_ligands, N_carbenes, bridge_atoms, neighbours = [], [], [], []
    charge_rules = np.zeros(len(mol.GetAtoms()), dtype=int)
    neighbours, metal_found, sanit_step = [], False, True
    for i, atom in enumerate(mol.GetAtoms()):
        # get the neighbours of metal atom and calculate the charge of metal center + ligands
        if atom.GetIdx() in args.metal_idx:
            # a sanitation step is needed to ensure that metals and ligands show correct valences
            if sanit_step:
                Chem.SanitizeMol(mol)
                sanit_step = False
            metal_found = True
            charge_idx = args.metal_idx.index(atom.GetIdx())
            neighbours = atom.GetNeighbors()
            charge_rules[i] = args.metal_oxi[charge_idx]
            for neighbour in neighbours:
                M_ligands.append(neighbour.GetIdx())
                if neighbour.GetTotalValence() == 4:
                    if neighbour.GetSymbol() in C_group:
                        # first, detects carbenes to adjust charge
                        carbene_like = False
                        bridge_ligand = False
                        for inside_neighbour in neighbour.GetNeighbors():
                            if inside_neighbour.GetSymbol() in N_group:
                                if inside_neighbour.GetTotalValence() == 4:
                                    for N_neighbour in inside_neighbour.GetNeighbors():
                                        # this option detects bridge ligands that connect two metals such as M--CN--M
                                        # we use I since the M is still represented as I at this point
                                        if N_neighbour.GetSymbol() == "I":
                                            bridge_ligand = True
                                            bridge_atoms.append(inside_neighbour.GetIdx())
                                    if not bridge_ligand:
                                        carbene_like = True
                                        N_carbenes.append(inside_neighbour.GetIdx())
                        if not carbene_like:
                            charge_rules[i] = charge_rules[i] - 1
                elif neighbour.GetTotalValence() == 3:
                    if neighbour.GetSymbol() in N_group and neighbour.GetFormalCharge() == 0:
                        charge_rules[i] = charge_rules[i] - 1
                elif neighbour.GetTotalValence() == 2:
                    # radical chalcogen atoms (i.e., Cu-OH(rad))
                    if neighbour.GetSymbol() in O_group and neighbour.GetFormalCharge() == 0 and len(neighbour.GetNeighbors()) == 2:
                        charge_rules[i] = charge_rules[i] - 1
                    # double bonded chalcogen atom (i.e., V=O)
                    elif neighbour.GetSymbol() in O_group and neighbour.GetFormalCharge() == 0 and len(neighbour.GetNeighbors()) == 1:
                        charge_rules[i] = charge_rules[i] - 2
                elif neighbour.GetTotalValence() == 1:
                    if neighbour.GetSymbol() in O_group:
                        charge_rules[i] = charge_rules[i] - 2
                    if neighbour.GetSymbol() in F_group:
                        charge_rules[i] = charge_rules[i] - 1

    # for charges not in the metal, neighbours or exceptions (i.e., C=N+ from carbenes or CN from bridge atoms)
    invalid_charged_atoms = M_ligands + N_carbenes + bridge_atoms + args.metal_idx
    for i, atom in enumerate(mol.GetAtoms()):
        if atom.GetIdx() not in invalid_charged_atoms:
            charge_rules[i] = atom.GetFormalCharge()

    charge = np.sum(charge_rules)
    if not metal_found:
        # for organic molecules when using a list containing organic and organometallics molecules mixed
        charge = Chem.GetFormalCharge(mol)

    return charge, metal_found


def substituted_mol(self, mol, checkI):
    """
    Returns a molecule object in which all metal atoms specified in args.metal_atoms
    are replaced by Iodine and the charge is set depending on the number of
    neighbors.

    """

    self.args.metal_idx = []
    self.args.complex_coord = []
    self.args.metal_sym = []

    for _ in self.args.metal_atoms:
        self.args.metal_idx.append(None)
        self.args.complex_coord.append(None)
        self.args.metal_sym.append(None)

    Neighbors2FormalCharge = dict()
    for i, j in zip(range(2, 9), range(-3, 4)):
        Neighbors2FormalCharge[i] = j

    for atom in mol.GetAtoms():
        symbol = atom.GetSymbol()
        if symbol in self.args.metal_atoms:
            self.args.metal_sym[self.args.metal_atoms.index(symbol)] = symbol
            self.args.metal_idx[self.args.metal_atoms.index(symbol)] = atom.GetIdx()
            self.args.complex_coord[self.args.metal_atoms.index(symbol)] = len(
                atom.GetNeighbors()
            )
            if checkI == "I":
                atom.SetAtomicNum(53)
                n_neighbors = len(atom.GetNeighbors())
                if n_neighbors > 1:
                    formal_charge = Neighbors2FormalCharge[n_neighbors]
                    atom.SetFormalCharge(formal_charge)

    return self.args.metal_idx, self.args.complex_coord, self.args.metal_sym


def set_metal_atomic_number(mol, metal_idx, metal_sym):
    """
    Changes the atomic number of the metal atoms using their indices.

    Parameters
    ----------
    mol : rdkit.Chem.Mol
        RDKit molecule object
    metal_idx : list
        sorted list that contains the indices of the metal atoms in the molecule
    metal_sym : list
        sorted list (same order as metal_idx) that contains the symbols of the metals in the molecule
    """

    for atom in mol.GetAtoms():
        if atom.GetIdx() in metal_idx:
            re_symbol = metal_sym[metal_idx.index(atom.GetIdx())]
            atomic_number = periodic_table().index(re_symbol)
            atom.SetAtomicNum(atomic_number)


def get_conf_RMS(mol1, mol2, c1, c2, heavy, max_matches_rmsd):
    """
    Takes in two rdkit.Chem.Mol objects and calculates the RMSD between them.
    (As side efect mol1 is left in the aligned state, if heavy is specified
    the side efect will not happen)

    Parameters
    ----------
    mol1 : rdkit.Chem.Mol
        Probe molecule
    mol2 : rdkit.Chem.Mol
        Target molecule. The probe is aligned to the target to compute the RMSD
    c1 : int
        Conformation of mol1 to use for the RMSD
    c2 : int
        Conformation of mol2 to use for the RMSD
    heavy : bool
        If True it will ignore the H atoms when computing the RMSD
    max_matches_rmsd : int
        Max number of matches found in a SubstructMatch()

    Returns
    -------
    float
        Returns the best RMSD found
    """

    if heavy:
        mol1 = RemoveHs(mol1)
        mol2 = RemoveHs(mol2)
    return GetBestRMS(mol1, mol2, c1, c2, maxMatches=max_matches_rmsd)


def command_line_args():
    """
    Load default and user-defined arguments specified through command lines. Arrguments are loaded as a dictionary
    """

    # First, create dictionary with user-defined arguments
    kwargs = {}
    available_args = ["help"]
    bool_args = [
        "csearch",
        "cmin",
        "qprep",
        "qcorr",
        "qdescp",
        "vismol",
        "heavyonly",
        "cregen",
        "lowest_only",
        "lowest_n",
        "chk",
    ]

    for arg in var_dict:
        if arg in bool_args:
            available_args.append(f"{arg}")
        else:
            available_args.append(f"{arg} =")

    try:
        opts, _ = getopt.getopt(sys.argv[1:], "h", available_args)
    except getopt.GetoptError as err:
        print(err)
        sys.exit()

    for arg, value in opts:
        if arg.find("--") > -1:
            arg_name = arg.split("--")[1].strip()
        elif arg.find("-") > -1:
            arg_name = arg.split("-")[1].strip()
        if arg_name in bool_args:
            value = True
        if value == "None":
            value = None
        if arg_name in ("h", "help"):
            print(f"o  AQME v {aqme_version} is installed correctly! For more information about the available options, see the documentation in https://github.com/jvalegre/aqme")
            sys.exit()
        else:
            # this "if" allows to use * to select multiple files in multiple OS
            if arg_name.lower() == 'files' and value.find('*') > -1:
                kwargs[arg_name] = glob.glob(value)
            else:
                # this converts the string parameters to lists
                if arg_name.lower() in ["files", "metal_oxi", "metal_atoms", "gen_atoms", "constraints_atoms", "constraints_dist", "constraints_angle", "constraints_dihedral", "atom_types", "cartesians", "nmr_atoms", "nmr_slope", "nmr_intercept"]:
                    if not isinstance(value, list):
                        try:
                            value = ast.literal_eval(value)
                        except (SyntaxError, ValueError):
                            pass
                kwargs[arg_name] = value

    # Second, load all the default variables as an "add_option" object
    args = load_variables(kwargs, "command")

    return args


def load_variables(kwargs, aqme_module, create_dat=True):
    """
    Load default and user-defined variables
    """

    # first, load default values and options manually added to the function
    self = set_options(kwargs)
    # this part loads variables from yaml files (if varfile is used)
    txt_yaml = ""
    if self.varfile is not None:
        self, txt_yaml = load_from_yaml(self)
    if aqme_module != "command":

        self.initial_dir = Path(os.getcwd())

        if not isinstance(self.files, list):
            self.w_dir_main = os.path.dirname(self.files)
            check_files = os.path.basename(self.files)
        elif len(self.files) != 0:
            self.w_dir_main = os.path.dirname(self.files[0])
        else:
            self.w_dir_main = os.getcwd()

        if (
            Path(f"{self.w_dir_main}").exists()
            and os.getcwd() not in f"{self.w_dir_main}"
        ):
            self.w_dir_main = Path(f"{os.getcwd()}/{self.w_dir_main}")
        else:
            self.w_dir_main = Path(self.w_dir_main)

        if self.isom_type is not None:
            if (
                Path(f"{self.isom_inputs}").exists()
                and os.getcwd() not in f"{self.isom_inputs}"
            ):
                self.isom_inputs = Path(f"{os.getcwd()}/{self.isom_inputs}")
            else:
                self.isom_inputs = Path(self.isom_inputs)

        error_setup = False

        if not self.w_dir_main.exists():
            txt_yaml += "\nx  The PATH specified as input in the w_dir_main option might be invalid! Using current working directory"
            error_setup = True

        if error_setup:
            self.w_dir_main = Path(os.getcwd())

        if not isinstance(self.files, list):
            if not isinstance(self.files, Mol):
                self.files = glob.glob(f"{self.w_dir_main}/{check_files}")
            else:
                self.files = [self.files]
        # this function is useful when PATH objects are given
        for i,file in enumerate(self.files):
            if not isinstance(file, Mol):
                self.files[i] = str(file)
            
        # start a log file to track the QCORR module
        if create_dat:
            logger_1, logger_2 = "AQME", "data"
            if aqme_module == "qcorr":
                # detects cycle of analysis (0 represents the starting point)
                self.round_num, self.resume_qcorr = check_run(self.w_dir_main)
                logger_1 = "QCORR-run"
                logger_2 = f"{str(self.round_num)}"

            elif aqme_module == "csearch":
                logger_1 = "CSEARCH"

            elif aqme_module == "cmin":
                logger_1 = "CMIN"

            elif aqme_module == "qprep":
                logger_1 = "QPREP"

            elif aqme_module == "qdescp":
                logger_1 = "QDESCP"

            elif aqme_module == "vismol":
                logger_1 = "VISMOL"

            if txt_yaml not in [
                "",
                f"\no  Importing AQME parameters from {self.varfile}",
                "\nx  The specified yaml file containing parameters was not found! Make sure that the valid params file is in the folder where you are running the code.\n",
            ]:
                self.log = Logger(self.initial_dir / logger_1, logger_2)
                self.log.write(txt_yaml)
                error_setup = True

            if not error_setup:
                if not self.command_line:
                    self.log = Logger(self.initial_dir / logger_1, logger_2)
                else:
                    # prevents errors when using command lines and running to remote directories
                    path_command = Path(f"{os.getcwd()}")
                    self.log = Logger(path_command / logger_1, logger_2)

                self.log.write(f"AQME v {aqme_version} {time_run} \nCitation: {aqme_ref}\n")

                if self.command_line:
                    self.log.write(f"Command line used in AQME: aqme {' '.join([str(elem) for elem in sys.argv[1:]])}\n")

            if error_setup:
                # this is added to avoid path problems in jupyter notebooks
                self.log.finalize()
                os.chdir(self.initial_dir)
                sys.exit()

    return self


def read_file(initial_dir, w_dir, file):
    """
    Reads through a file and retrieves a list with all the lines.
    """

    os.chdir(w_dir)
    outfile = open(file, "r")
    outlines = outfile.readlines()
    outfile.close()
    os.chdir(initial_dir)

    return outlines


def QM_coords(outlines, min_RMS, n_atoms, program, keywords_line):
    """
    Retrieves atom types and coordinates from QM output files
    """

    atom_types, cartesians, range_lines = [], [], []
    per_tab = periodic_table()
    count_RMS = -1

    if program == "gaussian":
        if "nosymm" in keywords_line.lower():
            target_ori = "Input orientation:"
        else:
            target_ori = "Standard orientation:"

        if min_RMS > -1:
            for i, line in enumerate(outlines):
                if line.find(target_ori) > -1:
                    count_RMS += 1
                if count_RMS == min_RMS:
                    range_lines = [i + 5, i + 5 + n_atoms]
                    break
        else:
            for i in reversed(range(len(outlines))):
                if outlines[i].find(target_ori) > -1:
                    range_lines = [i + 5, i + 5 + n_atoms]
                    break
        if len(range_lines) != 0:
            for i in range(range_lines[0], range_lines[1]):
                massno = int(outlines[i].split()[1])
                if massno < len(per_tab):
                    atom_symbol = per_tab[massno]
                else:
                    atom_symbol = "XX"
                atom_types.append(atom_symbol)
                cartesians.append(
                    [
                        float(outlines[i].split()[3]),
                        float(outlines[i].split()[4]),
                        float(outlines[i].split()[5]),
                    ]
                )

    return atom_types, cartesians


def cclib_atoms_coords(cclib_data):
    """
    Function to convert atomic numbers and coordinate arrays from cclib into
    a format compatible with QPREP.
    """

    atom_numbers = cclib_data["atoms"]["elements"]["number"]
    atom_types = []
    per_tab = periodic_table()
    for atom_n in atom_numbers:
        if atom_n < len(per_tab):
            atom_symbol = per_tab[atom_n]
        else:
            atom_symbol = "XX"
        atom_types.append(atom_symbol)

    cartesians_array = cclib_data["atoms"]["coords"]["3d"]
    cartesians = [
        cartesians_array[i : i + 3] for i in range(0, len(cartesians_array), 3)
    ]

    return atom_types, cartesians


def check_run(w_dir):
    """
    Determines the folder where input files are gonna be generated in QCORR.
    """

    if "failed" in w_dir.as_posix():
        resume_qcorr = True
        for folder in w_dir.as_posix().replace("\\", "/").split("/"):
            if "run_" in folder:
                folder_count = int(folder.split("_")[1]) + 1
    else:
        input_folder = w_dir.joinpath("failed/")
        resume_qcorr = False
        folder_count = 1

        if os.path.exists(input_folder):
            dir_list = os.listdir(input_folder)
            for folder in dir_list:
                if folder.find("run_") > -1:
                    folder_count += 1

    return folder_count, resume_qcorr


def read_xyz_charge_mult(file):
    """
    Reads charge and multiplicity from XYZ files. These parameters should be defined
    in the title lines as charge=X and mult=Y (i.e. FILENAME charge=1 mult=1 Eopt -129384.564)
    """

    charge_xyz, mult_xyz = None, None
    # read charge and mult from xyz files
    with open(file, "r") as F:
        lines = F.readlines()
    for line in lines:
        for keyword in line.strip().split():
            if keyword.lower().find("charge") > -1:
                charge_xyz = int(keyword.split("=")[1])
            elif keyword.lower().find("mult") > -1:
                mult_xyz = int(keyword.split("=")[1])
            elif charge_xyz is not None and mult_xyz is not None:
                break

    if charge_xyz is None:
        charge_xyz = 0
    if mult_xyz is None:
        mult_xyz = 1

    return charge_xyz, mult_xyz


def mol_from_sdf_or_mol_or_mol2(input_file, module):
    """
    mol object from SDF, MOL or MOL2 files
    """

    if module in ["qprep","cmin"]:
        # using sanitize=False to avoid reading problems
        mols = Chem.SDMolSupplier(input_file, removeHs=False, sanitize=False)
        return mols

    elif module == "csearch":

        # using sanitize=True in this case, which is recommended for RDKit calculations
        filename = os.path.splitext(input_file)[0]
        extension = os.path.splitext(input_file)[1]

        if extension.lower() == ".pdb":
            input_file = f'{input_file.split(".")[0]}.sdf'
            extension = ".sdf"

        if extension.lower() == ".sdf":
            mols = Chem.SDMolSupplier(input_file, removeHs=False)
        elif extension.lower() == ".mol":
            mols = [Chem.MolFromMolFile(input_file, removeHs=False)]
        elif extension.lower() == ".mol2":
            mols = [Chem.MolFromMol2File(input_file, removeHs=False)]

        IDs, charges, mults = [], [], []

        with open(input_file, "r") as F:
            lines = F.readlines()

        molecule_count = 0
        for i, line in enumerate(lines):
            if line.find(">  <ID>") > -1:
                ID = lines[i + 1].split()[0]
                IDs.append(ID)
            if line.find(">  <Real charge>") > -1:
                charge = lines[i + 1].split()[0]
                charges.append(charge)
            if line.find(">  <Mult>") > -1:
                mult = lines[i + 1].split()[0]
                mults.append(mult)
            if line.find("$$$$") > -1:
                molecule_count += 1
                if molecule_count != len(charges):
                    charges.append(0)
                if molecule_count != len(mults):
                    mults.append(1)

        suppl = []
        for i, mol in enumerate(mols):
            suppl.append(mol)

        if len(IDs) == 0:
            if len(suppl) > 1:
                for i in range(len(suppl)):
                    IDs.append(f"{filename}_{i+1}")
            else:
                IDs.append(filename)

        if len(charges) == 0:
            for i, mol in enumerate(mols):
                charges.append(Chem.GetFormalCharge(mol))
        if len(mults) == 0:
            for i, mol in enumerate(mols):
                NumRadicalElectrons = 0
                for Atom in mol.GetAtoms():
                    NumRadicalElectrons += Atom.GetNumRadicalElectrons()
                TotalElectronicSpin = NumRadicalElectrons / 2
                mult = int((2 * TotalElectronicSpin) + 1)
                mults.append(mult)

        return suppl, charges, mults, IDs



#!/usr/bin/python

#######################################################################
#             CREST functions from the CSEARCH module                 #
#######################################################################

from __future__ import print_function, absolute_import
import os
import glob
import numpy as np
from rdkit.Chem import AllChem as Chem
from rdkit.Chem import rdmolfiles
from rdkit import Geometry
import subprocess
import rdkit
from pathlib import Path
import shutil
from aqme.utils import read_file, run_command
from rdkit.Chem import rdMolTransforms


def atompairs(mol, atom1, atom2, constraints):
    active = []
    for x in constraints:
        active.append(x[:2])

    for i, x in enumerate(active):
        active[i] = [int(j) for j in x]

    pairs = []
    bonds = [(x.GetBeginAtomIdx(), x.GetEndAtomIdx()) for x in mol.GetBonds()]
    for [a, b] in bonds:
        if [a + 1, b + 1] not in active and [b + 1, a + 1] not in active:
            dist = round(rdMolTransforms.GetBondLength(mol.GetConformer(), a, b), 3)
            at_a, at_b = mol.GetAtoms()[a].GetSymbol(), mol.GetAtoms()[b].GetSymbol()
            if atom1 == "X" and atom2 == "X":
                pairs.append([float(a + 1), float(b + 1), dist])
            elif atom1 == "X" and atom2 == "H":
                if at_a == "H" or at_b == "H":
                    pairs.append([float(a + 1), float(b + 1), dist])
            else:
                if (at_a == atom1 and at_b == atom2) or (
                    at_a == atom2 and at_b == atom1
                ):
                    pairs.append([float(a + 1), float(b + 1), dist])
    return pairs


def get_constraint(mol, constraints):
    # constrained optimization with xtb
    xx_pairs = atompairs(mol, "X", "X", constraints)
    all_fix = []
    for x in constraints:
        all_fix.append(list(x))
    for x in xx_pairs:
        all_fix.append(x)
    return all_fix


def xyzall_2_xyz(xyzin, name):
    # converting multiple xyz to single
    command_run_1 = ["obabel", xyzin, "-oxyz", "-O" + name + "_conf_.xyz", "-m"]
    subprocess.run(command_run_1, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)


def xtb_opt_main(
    name,
    dup_data,
    dup_data_idx,
    self,
    charge,
    mult,
    constraints_atoms,
    constraints_dist,
    constraints_angle,
    constraints_dihedral,
    method_opt,
    complex_ts=False,
    mol=None,
    name_init=None
):

    """
    Run xTB using subprocess to perform CREST/CREGEN conformer sampling
    """

    name_no_path = name.replace("/", "\\").split("\\")[-1].split(".")[0]
    # where RDKit generates the files
    if self.args.destination is None:
        if method_opt == 'crest':
            csearch_dir = Path(self.args.w_dir_main) / "CSEARCH"
        elif method_opt == 'xtb':
            rdmolfiles.MolToXYZFile(mol, name + ".xyz")
            csearch_dir = Path(self.args.w_dir_main) / "CMIN"
    else:
        csearch_dir = Path(self.args.destination)
    if method_opt == 'crest':
        self.args.log.write(f"\no  Starting xTB pre-optimization before CREST sampling")
        dat_dir = csearch_dir / "crest_xyz"
        xyzin = f"{dat_dir}/{name_no_path}.xyz"
    elif method_opt == 'xtb':
        self.args.log.write(f"\no  Starting xTB optimization")
        dat_dir = csearch_dir / "xtb_xyz"
        xyzin = f"{dat_dir}/{name_no_path}_xtb.xyz"
    dat_dir.mkdir(exist_ok=True, parents=True)

    shutil.move(f"{name}.xyz", xyzin)

    os.environ["OMP_STACKSIZE"] = self.args.stacksize
    # to run xTB/CREST with more than 1 processor
    os.environ["OMP_NUM_THREADS"] = str(self.args.nprocs)
    cmin_valid = True

    os.chdir(dat_dir)
    
    # for systems that were created from 1D and 2D inputs (i.e. SMILES), this part includes two xTB
    # constrained optimizations to avoid geometry problems in noncovalent complexes and transition states

    # xTB optimization with all bonds frozen
    constrained_opt = False
    if len(constraints_atoms) > 0 or len(constraints_dist) > 0 or len(constraints_angle) > 0 or len(constraints_dihedral) > 0:
        constrained_opt = True
        complex_ts = True

    xyzoutxtb1 = str(dat_dir) + "/" + name_no_path + "_xtb1.xyz"
    if complex_ts:
        all_fix = get_constraint(mol, constraints_dist)

        _ = create_xcontrol(
            self.args,
            constraints_atoms,
            all_fix,
            [],
            [],
            xyzin,
            "constrain1.inp",
        )

        command1 = [
            "xtb",
            xyzin,
            "--opt",
            "--input",
            "constrain1.inp",
            "-c",
            str(charge),
            "--uhf",
            str(int(mult) - 1),
            "-P",
            str(self.args.nprocs),
        ]

        if self.args.xtb_keywords is not None:
            for keyword in self.args.xtb_keywords.split():
                command1.append(keyword)

        run_command(command1, "{}.out".format(xyzoutxtb1.split(".xyz")[0]))
        try:
            os.rename(str(dat_dir) + "/xtbopt.xyz", xyzoutxtb1)
        except FileNotFoundError:
            os.rename(str(dat_dir) + "/xtblast.xyz", xyzoutxtb1)

        # remove files that might interfere in subsequent calculations (i.e. wrong electron readings)
        for file in glob.glob('*') + glob.glob('*.*') + glob.glob('.*'):
            if os.path.exists(file):
                if file.find('_xtb2') == -1 and file.find('_xtb1') == -1 and file.find('.out') == -1:
                    os.remove(file)

        if constrained_opt:
            xyzoutxtb2 = str(dat_dir) + "/" + name_no_path + "_xtb2.xyz"
            # xTB optimization with the user-defined constraints
            _ = create_xcontrol(
                self.args,
                list(constraints_atoms),
                list(constraints_dist),
                list(constraints_angle),
                list(constraints_dihedral),
                xyzoutxtb1,
                "constrain2.inp",
            )

            command2 = [
                "xtb",
                xyzoutxtb1,
                "--opt",
                "--input",
                "constrain2.inp",
                "-c",
                str(charge),
                "--uhf",
                str(int(mult) - 1),
                "-P",
                str(self.args.nprocs),
            ]

            if self.args.xtb_keywords is not None:
                for keyword in self.args.xtb_keywords.split():
                    command2.append(keyword)

            run_command(command2, "{}.out".format(xyzoutxtb2.split(".xyz")[0]))
            try:
                os.rename(str(dat_dir) + "/xtbopt.xyz", xyzoutxtb2)
            except FileNotFoundError:
                os.rename(str(dat_dir) + "/xtblast.xyz", xyzoutxtb2)
        else:
            xyzoutxtb2 = xyzoutxtb1

    else:
        # Preoptimization with xTB to avoid issues from innacurate starting structures in CREST.
        # If you're dealing with a large system, increase the stack size
        try:
            command = [
                "xtb",
                xyzin,
                "--opt",
                "-c",
                str(charge),
                "--uhf",
                str(int(mult) - 1),
                "-P",
                str(self.args.nprocs), 
            ]

            if self.args.xtb_keywords is not None:
                for keyword in self.args.xtb_keywords.split():
                    command.append(keyword)

            run_command(command, f"{xyzin.split('.')[0]}_xtb1.out")
            os.rename(str(dat_dir) + "/xtbopt.xyz", xyzoutxtb1)
        except FileNotFoundError:
            self.args.log.write(f"\nx   There was an error during the xTB pre-optimization. This error is related to parallelization of xTB jobs and is normally observed when using metal complexes in some operative systems/OpenMP versions. AQME is switching to using one processor (nprocs=1).\n")
            self.args.nprocs = 1
            try:
                if self.args.xtb_keywords is None:
                    comm_xtb = f"export OMP_STACKSIZE={self.args.stacksize} && export OMP_NUM_THREADS={self.args.nprocs},1 \
                    && xtb {xyzin} --opt -c {charge} --uhf {int(mult) - 1} >> {xyzin.split('.')[0]}_xtb1.out"
                else:
                    comm_xtb = f"export OMP_STACKSIZE={self.args.stacksize} && export OMP_NUM_THREADS={self.args.nprocs},1 \
                    && xtb {xyzin} --opt -c {charge} --uhf {int(mult) - 1} {self.args.xtb_keywords} >> {xyzin.split('.')[0]}_xtb1.out"
                subprocess.call(comm_xtb, shell=False)
                os.rename(str(dat_dir) + "/xtbopt.xyz", xyzoutxtb1)
            except FileNotFoundError:
                if self.args.program.lower() == "crest":
                    self.args.log.write(f"\nx   There was another error during the xTB pre-optimization that could not be fixed. Trying CREST directly with no xTB preoptimization.\n")
                else:
                    self.args.log.write(f"\nx   There was another error during the xTB pre-optimization that could not be fixed (this molecule will be skipped).\n")
                cmin_valid = False
                mol_rd = None

        xyzoutxtb2 = xyzoutxtb1

    xyzoutall = str(dat_dir) + "/" + name_no_path + "_conformers.xyz"

    if self.args.program.lower() == "crest":
        self.args.log.write(f"\no  Starting CREST sampling")
        if constrained_opt:
            _ = create_xcontrol(
                self.args,
                list(constraints_atoms),
                list(constraints_dist),
                list(constraints_angle),
                list(constraints_dihedral),
                xyzoutxtb2,
                ".xcontrol.sample",
            )

        command = [
            "crest",
            xyzoutxtb2,
            "--chrg",
            str(charge),
            "--uhf",
            str(int(mult) - 1),
            "-T",
            str(self.args.nprocs),
            "--ewin",
            str(self.args.ewin_csearch),
        ]

        if constrained_opt:
            command.append("--cinp")
            command.append(".xcontrol.sample")

        if self.args.crest_keywords is not None:
            for keyword in self.args.crest_keywords.split():
                command.append(keyword)

        run_command(command, f"/{dat_dir}/{name_no_path}.out")

        # get number of n_atoms
        try:
            natoms = open("crest_best.xyz").readlines()[0].strip()
        except FileNotFoundError:
                self.args.log.write(f"\nx  CREST optimization failed! This might be caused by different reasons. For example, this might happen if you're using metal complexes without specifying any kind of template in the complex_type option (i.e. squareplanar).\n")

        if self.args.cregen and int(natoms) != 1:
            self.args.log.write(f"\no  Starting CREGEN sorting")
            command = ["crest", "crest_best.xyz", "--cregen", "crest_conformers.xyz"]

            if self.args.cregen_keywords is not None:
                for keyword in self.args.cregen_keywords.split():
                    command.append(keyword)

            run_command(command, f"{dat_dir}/{name_no_path}_cregen.out")

        try:
            if os.path.exists(str(dat_dir) + "/crest_clustered.xyz"):
                shutil.copy(str(dat_dir) + "/crest_clustered.xyz", xyzoutall)

            elif os.path.exists(str(dat_dir) + "/crest_ensemble.xyz"):
                shutil.copy(str(dat_dir) + "/crest_ensemble.xyz", xyzoutall)
            else:
                shutil.copy(str(dat_dir) + "/crest_conformers.xyz", xyzoutall)
        except FileNotFoundError:
            self.args.log.write("\nx   CREST conformer sampling failed! Please, try other options (i.e. include constrains, change the crest_keywords option, etc.)")
            cmin_valid = False

    if cmin_valid:
        if self.args.program.lower() == "crest":
            xyzall_2_xyz(xyzoutall, name_no_path)
            xyz_files = glob.glob(name_no_path + "_conf_*.xyz")
        if self.args.program.lower() == "xtb":
            xyz_files = [xyzoutxtb1]
        for _, file in enumerate(xyz_files):
            name_conf = file.split(".xyz")[0]
            command_xyz = ["obabel", "-ixyz", file, "-osdf", "-O" + name_conf + ".sdf"]
            subprocess.run(
                command_xyz, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
            )

        if self.args.program.lower() == "crest":
            sdwriter = Chem.SDWriter(str(f"{csearch_dir}/{name_no_path}.sdf"))
        sdf_files = glob.glob(name_no_path + "*.sdf")
        for file in sdf_files:
            mol = rdkit.Chem.SDMolSupplier(file, removeHs=False, sanitize=False)
            mol_rd = rdkit.Chem.RWMol(mol[0])
            if self.args.program.lower() == "xtb":
                # convert from hartree (default in xtb) to kcal
                energy_Eh = float(open(f'{file.split(".")[0]}.xyz', "r").readlines()[1].split()[1])
                energy_kcal = energy_Eh*627.5
                mol_rd.SetProp("_Name", name_init)
                os.remove(file)
            elif self.args.program.lower() == "crest":
                # convert from hartree (default in xtb) to kcal
                energy_Eh = float(open(file, "r").readlines()[0])
                energy_kcal = str(energy_Eh*627.5)
                mol_rd.SetProp("_Name", name_no_path)
                mol_rd.SetProp("Energy", energy_kcal)
                mol_rd.SetProp("Real charge", str(charge))
                mol_rd.SetProp("Mult", str(int(mult)))
                sdwriter.write(mol_rd)
                os.remove(file)
                os.remove(f'{file.split(".")[0]}.xyz')
    else:
        xyz_files = []

    # remove xTB/CREST files to avoid wrong readings of molecular information
    for file in glob.glob('*') + glob.glob('*.*') + glob.glob('.*'):
        if os.path.exists(file):
            if file.find('.out') == -1:
                if self.args.program.lower() == "xtb":
                    os.remove(file)
                elif self.args.program.lower() == "crest":
                    if file.find('_xtb2') == -1 and file.find('_xtb1') == -1 and file.find('.out') == -1:
                        try:
                            if file == 'crest_clustered.xyz':
                                os.rename('crest_clustered.xyz', f"{dat_dir}/{name_no_path}_clustered.xyz")
                            else:
                                os.remove(file)
                        except IsADirectoryError:
                            shutil.rmtree(file)

    os.chdir(self.args.w_dir_main)

    if method_opt == 'crest':
        dup_data.at[dup_data_idx, "crest-conformers"] = len(xyz_files)
        return 1

    if method_opt == 'xtb':
        return mol_rd, energy_kcal, cmin_valid


def create_xcontrol(
    args,
    constraints_atoms,
    constraints_dist,
    constraints_angle,
    constraints_dihedral,
    xyzin,
    name_constraint,
):
    """
    Function to create the .xcontrol.sample if constraints are defined
    """

    constrained_sampling = False

    unique_atoms = []
    for atom in constraints_atoms:
        if atom not in unique_atoms:
            unique_atoms.append(int(atom))
    for x in constraints_dist:
        for i in x[:2]:
            if i not in unique_atoms:
                unique_atoms.append(int(i))
    for x in constraints_angle:
        for i in x[:3]:
            if i not in unique_atoms:
                unique_atoms.append(int(i))
    for x in constraints_dihedral:
        for i in x[:4]:
            if i not in unique_atoms:
                unique_atoms.append(int(i))

    if len(unique_atoms) > 0:

        constrained_sampling = True

        # call --constrain just fo create a coord.ref file
        subprocess.run(
            ["crest", xyzin, "--constrain", "1"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

        # add the constraints part
        edited_xcontrol = "$constrain\n"

        if constraints_atoms != []:
            edited_xcontrol += "atoms: "
            edited_xcontrol += ",".join(str(atom_idx) for atom_idx in constraints_atoms)
            edited_xcontrol += "\n"

        for constraint_type in [
            constraints_dist,
            constraints_angle,
            constraints_dihedral,
        ]:
            if constraint_type != []:
                for const in constraint_type:
                    if constraint_type == constraints_dist:
                        edited_xcontrol += "distance: "
                        edited_xcontrol += (
                            ",".join(str(int(val)) for val in const[:2])
                            + ","
                            + str(const[2])
                        )
                    elif constraint_type == constraints_angle:
                        edited_xcontrol += "angle: "
                        edited_xcontrol += (
                            ",".join(str(int(val)) for val in const[:3])
                            + ","
                            + str(const[3])
                        )
                    elif constraint_type == constraints_dihedral:
                        edited_xcontrol += "dihedral: "
                        edited_xcontrol += (
                            ",".join(str(int(val)) for val in const[:4])
                            + ","
                            + str(const[4])
                        )
                    edited_xcontrol += "\n"

        edited_xcontrol += f"force constant={args.crest_force}\n"
        edited_xcontrol += "reference=coord.ref\n"

        # metadyn part
        if name_constraint == ".xcontrol.sample":
            outlines = read_file(os.getcwd(), os.getcwd(), xyzin)
            n_atoms = int(outlines[0])
            edited_xcontrol += "$metadyn\n"
            edited_xcontrol += "atoms: "
            for atom_idx in range(1, n_atoms + 1):
                if atom_idx not in unique_atoms:
                    if atom_idx == n_atoms:
                        edited_xcontrol += f"{atom_idx}\n"
                    else:
                        edited_xcontrol += f"{atom_idx},"
        edited_xcontrol += "$end\n"

        # write the file
        xcontrol_file = open(name_constraint, "w")
        xcontrol_file.write(edited_xcontrol)
        xcontrol_file.close()

    return constrained_sampling


def nci_ts_mol(
    smi,
    constraints_atoms,
    constraints_dist,
    constraints_angle,
    constraints_dihedral,
):
    if constraints_atoms is not None:
        constraints_atoms = [[float(y) for y in x] for x in constraints_atoms]
    if constraints_dist is not None:
        constraints_dist = [[float(y) for y in x] for x in constraints_dist]
        constraints_dist = np.array(constraints_dist)
    if constraints_angle is not None:
        constraints_angle = [[float(y) for y in x] for x in constraints_angle]
        constraints_angle = np.array(constraints_angle)
    if constraints_dihedral is not None:
        constraints_dihedral = [[float(y) for y in x] for x in constraints_dihedral]
        constraints_dihedral = np.array(constraints_dihedral)

    molsH = []
    mols = []
    for m in smi:
        mols.append(Chem.MolFromSmiles(m))
        molsH.append(Chem.AddHs(Chem.MolFromSmiles(m)))

    for m in molsH:
        Chem.EmbedMultipleConfs(m, numConfs=1)
    for m in mols:
        Chem.EmbedMultipleConfs(m, numConfs=1)

    coord = [0.0, 0.0, 5.0]
    molH = molsH[0]
    for _, fragment in enumerate(molsH[1:]):
        offset_3d = Geometry.Point3D(coord[0], coord[1], coord[2])
        molH = Chem.CombineMols(molH, fragment, offset_3d)
        coord[1] += 5
        Chem.SanitizeMol(molH)

    coord = [0.0, 0.0, 5.0]
    mol = mols[0]
    for _, fragment in enumerate(mols[1:]):
        offset_3d = Geometry.Point3D(coord[0], coord[1], coord[2])
        mol = Chem.CombineMols(mol, fragment, offset_3d)
        coord[1] += 5
        Chem.SanitizeMol(mol)
    mol = Chem.AddHs(mol)

    mol = Chem.ConstrainedEmbed(mol, molH)

    atom_map = []
    for atom in mol.GetAtoms():
        atom_map.append(atom.GetAtomMapNum())

    max_map = max(atom_map)
    for a in mol.GetAtoms():
        if a.GetSymbol() == "H":
            max_map += 1
            a.SetAtomMapNum(int(max_map))

    nconstraints_atoms = []
    if constraints_atoms is not None:
        for _, ele in enumerate(constraints_atoms):
            for atom in mol.GetAtoms():
                if ele == atom.GetAtomMapNum():
                    nconstraints_atoms.append(float(atom.GetIdx()) + 1)
        nconstraints_atoms = np.array(nconstraints_atoms)

    nconstraints_dist = []
    if constraints_dist is not None:
        for _, r in enumerate(constraints_dist):
            nr = []
            for _, ele in enumerate(r[:2]):
                for atom in mol.GetAtoms():
                    if ele == atom.GetAtomMapNum():
                        nr.append(float(atom.GetIdx()) + 1)
            nr.append(r[-1])
            nconstraints_dist.append(nr)
        nconstraints_dist = np.array(nconstraints_dist)

    nconstraints_angle = []
    if constraints_angle is not None:

        for _, r in enumerate(constraints_angle):
            nr = []
            for _, ele in enumerate(r[:3]):
                for atom in mol.GetAtoms():
                    if ele == atom.GetAtomMapNum():
                        nr.append(float(atom.GetIdx()) + 1)
            nr.append(r[-1])
            nconstraints_angle.append(nr)
        nconstraints_angle = np.array(nconstraints_angle)

    nconstraints_dihedral = []
    if constraints_dihedral is not None:
        for _, r in enumerate(constraints_dihedral):
            nr = []
            for _, ele in enumerate(r[:4]):
                for atom in mol.GetAtoms():
                    if ele == atom.GetAtomMapNum():
                        nr.append(float(atom.GetIdx()) + 1)
            nr.append(r[-1])
            nconstraints_dihedral.append(nr)
        nconstraints_dihedral = np.array(nconstraints_dihedral)

    return (
        mol,
        nconstraints_atoms,
        nconstraints_dist,
        nconstraints_angle,
        nconstraints_dihedral,
    )


#####################################################.
#        This file stores all the functions         #
#             used for filtering                    #
#####################################################.

# from functools import partial
# from rdkit.Chem import AllChem as Chem
# from rdkit.Chem import rdMolTransforms, Descriptors
from rdkit.Chem import Descriptors
from aqme.utils import periodic_table, get_conf_RMS

# Aux functions of the geometry filter
# def is_carbene_like(neighbours):
#     """
#     Takes a list of the metal neighbour atoms and returns True if they can be considered as if it were a carbene.

#     Parameters
#     ----------
#     neighbors : list
#         list of neighbour atoms of the metal.

#     Returns
#     -------
#     bool
#     """

#     carbene_like = False
#     N_group = ["N", "P", "As"]
#     for inside_neighbour in neighbours:
#         if inside_neighbour.GetSymbol() in N_group:
#             if inside_neighbour.GetTotalValence() == 4:
#                 carbene_like = True
#     return carbene_like


# def get_fragmented_versions(mol, atom_i, atom_j, metal_idx):
#     """
#     Creates two copies of the mol object where in the first one the atom_i and
#     in the second the atom_j are not bonded to the metal.

#     Parameters
#     ----------
#     mol : rdkit.Chem.Mol
#             The molecule to fragment.
#     atom_i : int
#             Idx of the first atom
#     atom_j : int
#             Idx of the second atom
#     metal_idx : int
#             Idx of the metal

#     Returns
#     -------
#     tuple
#             mol_i, mol_j
#     """

#     bond_i = mol.GetBondBetweenAtoms(atom_i, metal_idx)
#     mol_i = Chem.FragmentOnBonds(
#         mol, [bond_i.GetIdx()], addDummies=True, dummyLabels=[(atom_i, metal_idx)]
#     )
#     bond_j = mol.GetBondBetweenAtoms(atom_j, metal_idx)
#     mol_j = Chem.FragmentOnBonds(
#         mol, [bond_j.GetIdx()], addDummies=True, dummyLabels=[(atom_j, metal_idx)]
#     )
#     return mol_i, mol_j


# def is_around_angle(test, angle, offset):
#     """
#     Checks if a test angle is close to the angle or not.

#     Parameters
#     ----------
#     test : float
#             Angle to test in Degrees.
#     angle : float
#             Angle to compare in Degrees.
#     offset : float
#             Tolerance around 'angle' in degrees.

#     Returns
#     -------
#     bool
#             True if it is in the range [angle-offset,angle+offset].
#     """

#     return (angle - offset) <= test <= (angle + offset)


# def passes_Ir_bidentate_x3_rule_angle_requirements(
#     ligand_atoms, mol, metal_idx, offset
# ):
#     """
#     Checks if for complexes with Ph_Py ligands if the angles between the N
#     atoms are acceptable. Linear when there are 2 Ph_Py and not linear when
#     there are 3 Ph_Py Ligands.

#     Parameters
#     ----------
#     ligand_atoms : list
#             A list of tuples where the second value of each tuple is the Idx in the
#             provided mol of the N atom of that ligand.
#     mol : rdkit.Chem.Mol
#             The molcule object of the complex
#     metal_idx : int
#             Idx of the metal atom
#     offset : float
#             Angle error in degrees to consider or not an angle to be linear.

#     Returns
#     -------
#     bool
#             Whether it passes the angle requirements or not
#     """

#     is_lineal = partial(is_around_angle, angle=180, offset=offset)
#     get_angle = lambda i, j: rdMolTransforms.GetAngleDeg(mol, i, metal_idx, j)
#     passing = True

#     if len(ligand_atoms) == 3:  # For complexes with 3 Ph_Py ligands:
#         i, j, k = ligand_atoms
#         angles = [get_angle(a[1], b[1]) for a, b in [(i, j), (i, k), (j, k)]]
#         passing = not any([is_lineal(angle) for angle in angles])

#     if (
#         len(ligand_atoms) == 2
#     ):  # For complexes with 2 Ph_Py ligands + 1 ligand that is not Ph_Py
#         (i, i2), (j, j2) = ligand_atoms
#         angle = get_angle(i2, j2)
#         passing = is_lineal(angle)

#     return passing


# def passes_Ir_bidentate_x3_rule(mol, angle_off):
#     """
#     Checks if a mol containing an Ir metal complex passes the bidentate x3
#     rule or not.

#     Parameters
#     ----------
#     mol : rdkit.Chem.Mol
#             The molecule to be tested.
#     angle_off : float
#             Angle in degrees that acts as tolerance around 180º

#     Returns
#     -------
#     bool
#             True if it has passed the rule
#     """

#     bond_threshold = 2.60  # based on observation from DFT optimized geometries
#     ligand_links = []
#     atom_indexes = []

#     passing = True
#     # Finds the Ir atom and gets the atom types and indexes of all its neighbours
#     # the filter is compatible with molecules that do not contain Ir (always passing)
#     metal_idx = None
#     for atom in mol.GetAtoms():
#         if atom.GetAtomicNum() == 77:
#             metal_idx = atom.GetIdx()
#             for x in atom.GetNeighbors():
#                 ligand_links.append(x.GetSymbol())
#                 atom_indexes.append(x.GetIdx())

#     # I need to get the only 3D conformer generated in that mol object for rdMolTransforms
#     mol_conf = mol.GetConformer(0)
#     # This part will identify the pairs of C and N atoms that are part of the same Ph_Py ligand.
#     # The shape of the atom pairs is '[[C1_ATOM_NUMBER, N1_ATOM_NUMBER],[C2, N2],...]'.
#     # This information is required for the subsequent filtering process based on angles

#     # it filters off molecules that the SDF only detects 5 Ir neighbours
#     if len(atom_indexes) != 6:
#         if metal_idx is not None:
#             passing = False
#         return passing

#     # Filter by distance of the neighbour atoms and the metal
#     for atom_idx in atom_indexes:
#         bond_length = rdMolTransforms.GetBondLength(mol_conf, metal_idx, atom_idx)
#         if bond_length > bond_threshold:
#             return False

#     ligand_atoms = []
#     for atom_i, sym_i in zip(atom_indexes, ligand_links):
#         # This is a filter that excludes molecules that fell apart during DFT geometry
#         # optimization (i.e. a N atom from one of the ligands separated from Ir). The
#         # max distance allowed can be tuned in length_filter
#         for atom_j in atom_indexes:
#             # Avoid combinations of the same atom with itself
#             if atom_i != atom_j and sym_i == "C":
#                 # We know that the ligands never have 2 carbon atoms bonding the Ir atom except
#                 # for carbenes. We only use atom_indexes[i] for C atoms, and atom_indexes[j] for the potential
#                 # N atoms that are part of the same Ph_Py ligand
#                 neighbours = mol.GetAtoms()[atom_i].GetNeighbors()
#                 if not is_carbene_like(neighbours):
#                     # This part detects the Ir-C bond and breaks it, breaking the Ph_Py ring
#                     mol_i, mol_j = get_fragmented_versions(
#                         mol, atom_i, atom_j, metal_idx
#                     )
#                     # identify whether or not the initial 5-membered ring formed between [-Ir-C-C-C-N-] is broken when we break the Ir-C bond. This works
#                     # because Ph_Py units bind Ir in the same way always, through 1 C and 1 N that are in the same position, forming a 5-membered ring.
#                     # If this ring is broken, atom_indexes[j] will not be part of a 5-membered ring (atom.IsInRingSize(5) == False) which means that
#                     # this atom was initially inside the same ligand as the parent C of atom_indexes[i])
#                     if not mol_i.GetAtomWithIdx(atom_i).IsInRingSize(5):
#                         # doing backwards as well eg. Ir N bond
#                         if not mol_j.GetAtomWithIdx(atom_i).IsInRingSize(5):
#                             ligand_atoms.append([atom_i, atom_j])
#                             break
#                     else:
#                         if not mol_i.GetAtomWithIdx(atom_j).IsInRingSize(5):
#                             if mol.GetAtomWithIdx(atom_j).IsInRingSize(5):
#                                 ligand_atoms.append([atom_i, atom_j])
#                                 break
#     # Check Angles
#     passing = passing and passes_Ir_bidentate_x3_rule_angle_requirements(
#         ligand_atoms, mol_conf, metal_idx, angle_off
#     )
#     return passing


# # Main API of the geometry filter
# def geom_rules_output(mol, args, log, file, print_error_geom_rules):
#     """
#     returns if a mol object passes all the discarding rules or not.

#     Parameters
#     ----------
#     mol : rdkit.Chem.Mol
#             molecule to be tested.
#     args : argparse.args
#             [description]
#     log : Logger
#             [description]
#     file : str
#             Only used to write to the log
#     print_error_geom_rules : bool
#             Controls extra writing to the log.

#     Returns
#     -------
#     bool
#             If True, it means that it is in accordance with the rules
#     """

#     passing = True
#     for rule in args.geom_rules:
#         if rule == "Ir_bidentate_x3":
#             passing = passes_Ir_bidentate_x3_rule(mol, args.angle_off)
#         else:
#             var = rule.split(",")
#             if len(var) < 2:
#                 log.write("x  The geom_rules parameter(s) was not correctly defined, this filter will be turned off")
#                 return True
#             atoms_filter = var[0].split("-")
#             angle_rules = int(var[1])
#             # the elements of this initial list will be replaced by the corresponding atom id numebrs
#             atom_idx = ["ATOM1", "ATOM2", "ATOM3"]

#             find_angle = 0
#             incompatibility_found = False
#             for atom in mol.GetAtoms():
#                 # count matches
#                 neigh_count_first = 0
#                 neigh_count_second = 0
#                 sym_0, sym_1, sym_2 = atoms_filter
#                 # Finds the metal atom and gets the atom types and indexes of all its neighbours
#                 if atom.GetSymbol() == sym_1:
#                     # idx of the central atom
#                     atom_idx[1] = atom.GetIdx()
#                     for x in atom.GetNeighbors():
#                         sym = x.GetSymbol()
#                         if sym == sym_0 or sym == sym_2:
#                             # this ensures that both neighbours are used
#                             if sym == sym_0 and sym == sym_2:
#                                 if neigh_count_first <= neigh_count_second:
#                                     neigh_count_first += 1
#                                     atom_idx[0] = x.GetIdx()
#                                 else:
#                                     neigh_count_second += 1
#                                     atom_idx[2] = x.GetIdx()
#                             elif sym == sym_0:
#                                 neigh_count_first += 1
#                                 atom_idx[0] = x.GetIdx()
#                             elif sym == sym_2:
#                                 neigh_count_second += 1
#                                 atom_idx[2] = x.GetIdx()
#                             # count matches
#                             matches = neigh_count_first + neigh_count_second
#                             if matches > 2:
#                                 if not print_error_geom_rules:
#                                     log.write(f"x  There are multiple options in geom_rules for {file}, this filter will be turned off")
#                                     incompatibility_found = True
#                                     break
#                     if neigh_count_first == 1 and neigh_count_second == 1:
#                         find_angle += 1
#             if find_angle == 0 and not incompatibility_found:
#                 if not print_error_geom_rules:
#                     log.write(f"x  No angles matching the description from geom_rules in {file}, this filter will be turned off")
#             elif find_angle > 1:
#                 log.write(f"x  {file} contain more than one atom that meets the geom_rules criteria, this filter will be turned off")
#             elif find_angle == 1:
#                 # Retrieve the only 3D conformer generated in that mol object for rdMolTransforms
#                 mol_conf = mol.GetConformer(0)
#                 # Calculate the angle between the 3 elements
#                 angle = rdMolTransforms.GetAngleDeg(
#                     mol_conf, atom_idx[0], atom_idx[1], atom_idx[2]
#                 )
#                 passing = not is_around_angle(angle, angle_rules, args.angle_off)
#             if not passing:
#                 break
#     return passing

# Mol filters
def filters(mol, log, molwt_cutoff):
    """
    Applies some basic filters (molwt, salts[currently off], weird atom symbols)
    that only require SMILES data from a compound and returns if the molecule
    passes the filters or not.
    """

    # Filter 1
    molwt_cutoff = float(molwt_cutoff)
    if Descriptors.MolWt(mol) >= molwt_cutoff and molwt_cutoff > 0:
        log.write(f"x   Skipping this molecule as total molar mass > {molwt_cutoff}")
        return False

    # Filter 2
    symbols = [atom.GetSymbol() for atom in mol.GetAtoms()]

    unknown_atoms = list(set(symbols) - set(periodic_table()))
    if unknown_atoms:
        log.write(f" Exiting as atoms [{','.join(unknown_atoms)}] are not in the periodic table")
        return False

    # Passed
    return True


def ewin_filter(
    sorted_all_cids,
    cenergy,
    dup_data,
    dup_data_idx,
    calc_type,
    energy_window,
):
    """
    Given a sorted list of Compound Ids and a sorted list of their energies
    it discards all compound Ids that have an energy higher than the
    args.ewin_csearch with respect to the lowest one.

    Parameters
    ----------
    sorted_all_cids : list
            [description]
    cenergy : list
            [description]
    args : argparse.args
            [description]
    dup_data : pd.Dataframe
            [description]
    dup_data_idx : pd.Dataframe?
            [description]
    calc_type : str
            A string that points towards the column of the dataframe that should
            be filled with the number of duplicates. The current choices are:
            ['rdkit','summ','ani','xtb']
    energy_window : float
            Minimum energy difference with respect to the lowest compound
            discard a compound.

    Returns
    -------
    list
            list of cids accepted
    """

    sortedcids = []
    count = 0
    energy_window = float(energy_window)

    cenergy_min = cenergy[sorted_all_cids[0]]
    # Filter by Energy Window
    for cid in sorted_all_cids:
        if abs(cenergy[cid] - cenergy_min) < energy_window:
            sortedcids.append(cid)
        else:
            count += 1

    if calc_type == "rdkit":
        key = "RDKit"
    elif calc_type == "summ":
        key = "summ"
    elif calc_type == "ani":
        key = "ANI"
    elif calc_type == "xtb":
        key = "xTB"

    # Write it
    dup_data.at[dup_data_idx, f"{key}-energy-window"] = count

    return sortedcids


def pre_E_filter(
    sortedcids, cenergy, dup_data, dup_data_idx, calc_type, threshold
):
    """
    This filter selects the first compound that it finds with energy an energy
    difference higher or equal to the threshold with respect to the previously
    admitted compounds. (Thought as filter for rdkit)

    Parameters
    ----------
    sortedcids : list or pd.Dataframe?
            List of compound Ids.
    cenergy : list or pd.Dataframe?
            list of compound energies
    dup_data : pd.Dataframe
            [description]
    dup_data_idx : pd.Dataframe?
            [description]
    calc_type : str
            A string that points towards the column of the dataframe that should
            be filled with the number of duplicates. The current choices are:
            ['rdkit','summ','ani','xtb']
    threshold : float
            Minimum energy difference to consider two compounds as different.
            (kcal/mol)

    Returns
    -------
    list
            list of accepted compound Ids
    """

    selectedcids_initial = []
    eng_dup = 0
    threshold = float(threshold)

    # Add the first one
    selectedcids_initial.append(sortedcids[0])
    for conf in sortedcids[1:]:
        is_unique = True
        # check rmsd
        for seenconf in selectedcids_initial:
            E_diff = abs(cenergy[conf] - cenergy[seenconf])  # in kcal/mol
            if E_diff < threshold:
                eng_dup += 1
                is_unique = False
                break
        if is_unique:
            if conf not in selectedcids_initial:
                selectedcids_initial.append(conf)

    if calc_type == "rdkit":
        column = "RDKit-initial_energy_threshold"
    elif calc_type == "summ":
        column = "summ-initial_energy_threshold"
    # if calc_type == 'fullmonte': column = 'FullMonte-initial_energy_threshold'
    elif calc_type == "ani":
        column = "ANI-initial_energy_threshold"
    elif calc_type == "xtb":
        column = "xTB-initial_energy_threshold"
    else:
        column = ""

    if column:
        dup_data.at[dup_data_idx, column] = eng_dup
    return selectedcids_initial


def RMSD_and_E_filter(
    outmols, selectedcids_initial, cenergy, args, dup_data, dup_data_idx, calc_type
):
    """
    This filter selects the first compound that it finds with energy an energy
    difference lower than the threshold with a higher than the threshold rms
    with respect to the nearest (in energy) accepted compound.
    """

    selectedcids = []
    eng_rms_dup = 0
    selectedcids.append(selectedcids_initial[0])
    energy_threshold = float(args.energy_threshold)
    rms_threshold = float(args.rms_threshold)
    max_matches_rmsd = int(args.max_matches_rmsd)

    for _,conf in enumerate(selectedcids_initial[1:]):
        # This keeps track of whether or not your conformer is unique
        excluded_conf = False

        # check energy and rmsd
        for seenconf in selectedcids:
            E_diff = abs(cenergy[conf] - cenergy[seenconf])  # in kcal/mol
            if E_diff < energy_threshold:
                if calc_type == "rdkit":
                    rms = get_conf_RMS(
                        outmols[seenconf],
                        outmols[conf],
                        seenconf,
                        conf,
                        args.heavyonly,
                        max_matches_rmsd
                    )
                # elif calc_type == 'summ' or calc_type == 'fullmonte' or calc_type =='xtb' or calc_type =='ani':
                elif calc_type == "summ" or calc_type == "xtb" or calc_type == "ani":
                    rms = get_conf_RMS(
                        outmols[conf],
                        outmols[seenconf],
                        -1,
                        -1,
                        args.heavyonly,
                        max_matches_rmsd
                    )
                if rms < rms_threshold:
                    excluded_conf = True
                    eng_rms_dup += 1
                    break

        if not excluded_conf:
            if conf not in selectedcids:
                selectedcids.append(conf)

    # Write the found duplicates:
    if calc_type == "rdkit":
        key = "RDKit"
    elif calc_type == "summ":
        key = "summ"
    elif calc_type == "ani":
        key = "ANI"
    elif calc_type == "xtb":
        key = "xTB"
    else:
        key = ""

    if key:
        duplicates_column = f"{key}-RMSD-and-energy-duplicates"
        uniques_column = f"{key}-Unique-conformers"
        dup_data.at[dup_data_idx, duplicates_column] = eng_rms_dup
        dup_data.at[dup_data_idx, uniques_column] = len(selectedcids)

    return selectedcids



#####################################################.
#       This file stores functions related to       #
#          metal templates used in CSEARCH          #
#####################################################.

import sys
from pathlib import Path
from pkg_resources import resource_filename
from rdkit.Chem import AllChem as Chem
from rdkit.Chem import rdDistGeom, rdMolAlign
from aqme.utils import get_conf_RMS

TEMPLATES_PATH = Path(resource_filename("aqme", "templates"))


def template_embed(self, mol, complex_type, metal_idx, maxsteps, heavyonly, maxmatches, name):
    """
    Wrapper function to select automatically the appropiate embedding function
    depending on the number of neighbours of the metal center.
    """

    embed_functions = dict()
    embed_functions[2] = two_embed
    embed_functions[3] = three_embed
    embed_functions[4] = four_embed
    embed_functions[5] = five_embed

    template = load_template(complex_type, self.args.log)

    # Generate the embeddings
    neighbours = calc_neighbours(mol, metal_idx)
    embed = embed_functions[len(neighbours)]
    items = embed(mol, template, neighbours, name, maxsteps, self.args.log)

    # Filter the results
    molecules = items[0]
    if len(molecules) > 1:
        ignored = []
        for i, mol_filter in enumerate(molecules):
            has_big_rmsd = filter_template_mol(
                mol_filter, molecules, heavyonly, maxmatches
            )
            if has_big_rmsd:
                ignored.append(i)
        items = [item for i, item in enumerate(items) if i not in ignored]
    return items


def template_embed_optimize(target, template, maxsteps, log):
    """
    Embeds a new conformation into a molecule, optimizes it using UFF and
    realigns it.

    Parameters
    ----------
    target : rdkit.Chem.Mol
        Molecule where you want to embed the new conformation
    mol_template : rdkit.Chem.Mol?
        Template molecule to identify the core of the molecule that will have
        its distances frozen in the optimization.
    maxsteps : int
        Number of maximum optimization steps in RDKit.
    log : Logger
        [description]

    Returns
    -------
    molecule, coordMap, algMap, conf_id
        Molecule embedded, mapping to the atom instances,
        list of tuples with position and id and int with the conformer id.
    """

    seed = -1
    force_constant = 10000

    coord_map, alg_map = get_mappings(target, template, conformer_id=-1)

    # add H's to molecule
    molecule = Chem.AddHs(target)

    conf_id = rdDistGeom.EmbedMolecule(molecule, coordMap=coord_map, randomSeed=seed)

    if conf_id < 0:
        log.write("Could not embed molecule.")
        return molecule, None, None, conf_id

    forcefield = Chem.UFFGetMoleculeForceField(molecule, confId=conf_id)

    constraints = get_distance_constrains(coord_map)
    for i, j, d in constraints:
        forcefield.AddDistanceConstraint(i, j, d, d, force_constant)
    forcefield.Initialize()
    forcefield.Minimize(maxIts=maxsteps)
    # rotate the embedded conformation onto the core_mol:
    rdMolAlign.AlignMol(molecule, template, atomMap=alg_map, reflect=True, maxIters=100)

    return molecule, coord_map, alg_map, conf_id


def filter_template_mol(molecule_new, mol_objects, heavyonly, max_matches):
    """
    Returns if a mol object should be kept or not.

    Parameters
    ----------
    molecule_new : [type]
        [description]
    mol_objects : [type]
        [description]
    heavyonly : bool
        If True only non-H atoms are considered for the RMS calculation
    max_matches : int
        Maximum number of matches in the RMSD?

    Returns
    -------
    bool
        Returns True when the molecule should be kept and False when it should be discarded.
    """

    if not mol_objects:
        return True

    # check if molecule also exixts in the mol_objects
    for mol in mol_objects:
        rms = get_conf_RMS(mol, molecule_new, -1, -1, heavyonly, max_matches)
        if rms < 0.5:
            return False
    return True


def get_mappings(molecule, template, conformer_id=-1):
    match = molecule.GetSubstructMatch(template)
    conformer = template.GetConformer(conformer_id)
    coordMap = {}
    algMap = []
    for i, atomidx in enumerate(match):
        algMap.append((atomidx, i))
        coordMap[atomidx] = conformer.GetAtomPosition(i)
    return coordMap, algMap


def load_template(complex_type, log):
    """
    Checks if the templates are reachable and if so returns the molecule object.

    Returns
    -------
    rdkit.Chem.Mol
        The molecule file of the corresponding template.
    """

    type2template = dict()
    type2template["squareplanar"] = "template-4-and-5.sdf"
    type2template["squarepyramidal"] = "template-4-and-5.sdf"
    type2template["linear"] = "template-2.sdf"
    type2template["trigonalplanar"] = "template-3.sdf"

    folder = TEMPLATES_PATH

    if not folder.exists():
        log.write("x  The templates folder was not found, probably due to a problem while installing AQME")
        log.finalize()
        sys.exit()

    file_template = folder / Path(type2template[complex_type])
    templates = Chem.SDMolSupplier(str(file_template))
    template = templates[-1]

    return template


def calc_neighbours(molecule, metals_idx):
    """
    Changes the atomic number (and charge) of the first metal found
    and returns its neighbours, the number of neighbours and the idx of the
    metal.

    Parameters
    ----------
    molecule : rdkit.Chem.Mol
        Mol object
    metals_idx : list
        List containing the Idx of the metals. The first match is the only one considered.

    Returns
    -------
    neighbours : list
        List of neighbour atoms
    """

    # depending on the amount of neighbours, use Si or I atoms to fit the templates
    bonds2AtNum = dict()
    bonds2AtNum[5] = 14
    bonds2AtNum[4] = 14
    bonds2AtNum[3] = 53
    bonds2AtNum[2] = 53

    for atom in molecule.GetAtoms():
        idx = atom.GetIdx()
        if idx in metals_idx:
            neighbours = atom.GetNeighbors()
            # in case metals are used with different bonds (i.e., Cu2+ and CuL2)
            n_bonds = len(atom.GetBonds())
            AtNum = bonds2AtNum[n_bonds]
            atom.SetAtomicNum(AtNum)
            if n_bonds == 5:
                atom.SetFormalCharge(1)
            return neighbours
    return []


def check_metal_neigh(mol, complex_type, metal_idx_ind, log, valid_template):
    """
    Checks if the metal and the template contain the same number of ligands.

    Parameters
    ----------
    mol : rdkit.Chem.Mol
        Mol object
    complex_type : str
        Type of template to be used (i.e., squareplanar)
    metals_idx : list
        List containing the Idx of the metals

    Returns
    -------
    valid_template : bool
        Whether the complexes are compatible with the template selected
    """

    if complex_type == "linear":
        expect_neigh = 2
    elif complex_type == "trigonalplanar":
        expect_neigh = 3
    elif complex_type == "squareplanar":
        expect_neigh = 4
    elif complex_type == "squarepyramidal":
        expect_neigh = 5
    metal_atom = mol.GetAtoms()[metal_idx_ind]
    metal_neigh = metal_atom.GetNeighbors()
    if len(metal_neigh) != expect_neigh:
        log.write(f"x  The number of neighbours of the metal ({len(metal_neigh)}) does not match the number of expected neighbours for the template selected ({complex_type}). No templates will be applied to this system.")
        valid_template = False

    return valid_template


def get_distance_constrains(coordMap):
    atom_idxs = list(coordMap.keys())
    constrains = []
    for k, i in enumerate(atom_idxs):
        for j in atom_idxs[k + 1 :]:
            d = coordMap[i].Distance(coordMap[j])
            constrains.append((i, j, d))
    return constrains


# Decorators for documentation
def doc_parameters(f):
    """
    Decorator that adds the "Parameters" section at the end of the
    docstrings of the decorated function. Care to use this decorator 'below' the
    doc_returns decorator.

    Parameters
    ----------
    f : function
        Function to decorate.

    Returns
    -------
    function
        Returns the same function with the docstring modified.
    """

    description = f.__doc__
    parameters = [
        ("molecule", "rdkit.Chem.Mol", "Molecule to be embedded "),
        ("template", "rdkit.Chem.Mol", "Template molecule to do the embedding"),
        ("neighbours", "list", "Idx of the atoms neighbouring the metal"),
        ("name_input", "str", "Base name for the embedded molecules"),
        ("maxsteps", "int", "Maximum number of optimization steps run with rdkit."),
        ("log", "Logger", "[description]"),
    ]
    item_fmt = "{} : {}\n    {}".format
    params_txt = "\n".join([item_fmt(*items) for items in parameters])
    f.__doc__ = f"{description}\nParameters\n----------\n{params_txt}\n"
    return f


def doc_returns(f):
    """
    Decorator that adds the "Returns" section at the end of the
    docstrings of the decorated function.

    Parameters
    ----------
    f : function
        Function to decorate.

    Returns
    -------
    function
        Returns the same function with the docstring modified.
    """

    description = f.__doc__
    item_fmt = "{} : {}\n    {}".format
    outputs = [
        ("mol_objects", "list", "Embedded molecules."),
        ("name_returns", "list", "Names for the embedded molecules"),
        (
            "coord_maps",
            "list",
            "Mappings to the Idx of the atoms afected by the embedding",
        ),
        ("alg_maps", "list", "Mappings to the Idx of the core to do alignments"),
        ("mol_templates", "list", "List of template molecules used"),
    ]
    outs_txt = "\n".join([item_fmt(*items) for items in outputs])
    f.__doc__ = f"{description}\nReturns\n-------\n{outs_txt}\n"
    return f


# Embedding functions
@doc_returns
@doc_parameters
def two_embed(molecule, template, neighbours, name, maxsteps, log):
    """
    Embedding function for linear geometries. Requires 'linear.sdf' template.
    """

    template.GetAtomWithIdx(0).SetAtomicNum(neighbours[0].GetAtomicNum())
    template.GetAtomWithIdx(1).SetAtomicNum(neighbours[1].GetAtomicNum())
    template.GetAtomWithIdx(2).SetAtomicNum(53)

    # assigning and embedding onto the core
    mol_obj, coord_map, alg_map, ci = template_embed_optimize(
        molecule, template, maxsteps, log
    )
    if ci >= 0:  # writing to mol_object file
        return [mol_obj], [name], [coord_map], [alg_map], [template]

    return [], [], [], [], []


@doc_returns
@doc_parameters
def three_embed(molecule, template, neighbours, name, maxsteps, log):
    """
    Embedding function for trigonal planar geometry. Requires
    'trigonalplanar.sdf' template.
    """

    template.GetAtomWithIdx(0).SetAtomicNum(53)
    template.GetAtomWithIdx(1).SetAtomicNum(neighbours[0].GetAtomicNum())
    template.GetAtomWithIdx(2).SetAtomicNum(neighbours[1].GetAtomicNum())
    template.GetAtomWithIdx(3).SetAtomicNum(neighbours[2].GetAtomicNum())

    # assigning and embedding onto the core
    mol_obj, coord_map, alg_map, conf_id = template_embed_optimize(
        molecule, template, maxsteps, log
    )
    if conf_id >= 0:  # writing to mol_object file
        return [mol_obj], [name], [coord_map], [alg_map], [template]

    return [], [], [], [], []


@doc_returns
@doc_parameters
def four_embed(molecule, template, neighbours, name, maxsteps, log):
    """
    Embedding function for squareplanar geometry. Requires 'template-4-and-5.sdf'
    template. Attempts 3 embeddings.
    """

    mol_objects = []
    name_return = []
    coord_maps = []
    alg_maps = []
    mol_templates = []

    # Remove F atoms from the template
    for atom in template.GetAtoms():
        if atom.GetSymbol() == "F":
            template = Chem.RWMol(template)
            template.RemoveAtom(atom.GetIdx())
    template = template.GetMol()

    # three cases for square planar
    atn0 = neighbours[0].GetAtomicNum()
    atn1 = neighbours[1].GetAtomicNum()
    atn2 = neighbours[2].GetAtomicNum()
    atn3 = neighbours[3].GetAtomicNum()
    replacements_list = [
        (atn0, atn1, atn2, atn3, 14),
        (atn0, atn2, atn3, atn1, 14),
        (atn0, atn3, atn1, atn2, 14),
    ]

    for i, replacements in enumerate(replacements_list):

        # Create a copy of the mol object
        mol = Chem.Mol(template)

        # Assign atomic numbers to neighbour atoms
        for idx, atn in enumerate(replacements):
            mol.GetAtomWithIdx(idx).SetAtomicNum(atn)

        # embedding of the molecule onto the core
        mol_obj, coord_map, alg_map, conf_id = template_embed_optimize(
            molecule, mol, maxsteps, log
        )

        if conf_id >= 0:
            name_out = f"{name.split()[0]}_{i}"
            mol_objects.append(mol_obj)
            name_return.append(name_out)
            coord_maps.append(coord_map)
            alg_maps.append(alg_map)
            mol_templates.append(mol)

    return mol_objects, name_return, coord_maps, alg_maps, mol_templates


@doc_returns
@doc_parameters
def five_embed(molecule, template, neighbours, name, maxsteps, log):
    """
    Embedding function for squarepyramidal geometry. Requires
    'template-4-and-5.sdf' template. Attempts 15 embeddings.
    """

    mol_objects = []
    name_return = []
    coord_maps = []
    alg_maps = []
    mol_templates = []
    counter = 0
    atomic_numbers = []
    for _,atom in enumerate(neighbours):
        atomic_numbers.append(atom.GetAtomicNum())
    replacements = [
        [4, 0, 1, 2, 3],
        [4, 0, 2, 3, 1],
        [4, 0, 3, 1, 2],
        [3, 0, 1, 2, 4],
        [3, 0, 2, 4, 1],
        [3, 0, 4, 1, 2],
        [2, 0, 1, 4, 3],
        [2, 0, 4, 3, 1],
        [2, 0, 4, 1, 3],
        [1, 0, 4, 2, 3],
        [1, 0, 2, 3, 4],
        [1, 0, 3, 4, 2],
        [0, 4, 1, 2, 3],
        [0, 4, 2, 3, 1],
        [0, 4, 3, 1, 2],
    ]
    for replacement in replacements:
        at0, at1, at2, at3, at4 = [atomic_numbers[r] for r in replacement]
        template.GetAtomWithIdx(0).SetAtomicNum(at0)
        template.GetAtomWithIdx(1).SetAtomicNum(at1)
        template.GetAtomWithIdx(2).SetAtomicNum(at2)
        template.GetAtomWithIdx(3).SetAtomicNum(at3)
        template.GetAtomWithIdx(4).SetAtomicNum(at4)
        template.GetAtomWithIdx(5).SetAtomicNum(14)
        template.GetAtomWithIdx(5).SetFormalCharge(1)

        # assigning and embedding onto the core
        mol_obj, coord_map, alg_map, conf_id = template_embed_optimize(
            molecule, template, maxsteps, log
        )
        if conf_id >= 0:
            name_out = f"{name}_{counter}"
            mol_objects.append(mol_obj)
            name_return.append(name_out)
            coord_maps.append(coord_map)
            alg_maps.append(alg_map)
            mol_templates.append(template)
            counter += 1
    return mol_objects, name_return, coord_maps, alg_maps, mol_templates



######################################################.
#        This file stores functions related          #
#               to the QCORR module                  #
######################################################.

import os
import glob
import pandas as pd
import json
import cclib
from pathlib import Path
from aqme.utils import move_file, read_file, Logger
import numpy as np

# Bondi VDW radii in Angstrom
bondi = {
    "H": 1.09,
    "He": 1.40,
    "Li": 1.81,
    "Be": 1.53,
    "B": 1.92,
    "C": 1.70,
    "N": 1.55,
    "O": 1.52,
    "F": 1.47,
    "Ne": 1.54,
    "Na": 2.27,
    "Mg": 1.73,
    "Al": 1.84,
    "Si": 2.10,
    "P": 1.80,
    "S": 1.80,
    "Cl": 1.75,
    "Ar": 1.88,
    "K": 2.75,
    "Ca": 2.31,
    "Ni": 1.63,
    "Cu": 1.40,
    "Zn": 1.39,
    "Ga": 1.87,
    "Ge": 2.11,
    "As": 1.85,
    "Se": 1.90,
    "Br": 1.83,
    "Kr": 2.02,
    "Rb": 3.03,
    "Sr": 2.49,
    "Pd": 1.63,
    "Ag": 1.72,
    "Cd": 1.58,
    "In": 1.93,
    "Sn": 2.17,
    "Sb": 2.06,
    "Te": 2.06,
    "I": 1.98,
    "Xe": 2.16,
    "Cs": 3.43,
    "Ba": 2.68,
    "Pt": 1.72,
    "Au": 1.66,
    "Hg": 1.55,
    "Tl": 1.96,
    "Pb": 2.02,
    "Bi": 2.07,
    "Po": 1.97,
    "At": 2.02,
    "Rn": 2.20,
    "Fr": 3.48,
    "Ra": 2.83,
    "U": 1.86,
}

# covalent radii in Angstrom (taken from Pyykko and Atsumi, Chem. Eur. J. 15, 2009, 188-197)
rcov = {
    "H": 0.32,
    "He": 0.46,
    "Li": 1.33,
    "Be": 1.02,
    "B": 0.85,
    "C": 0.75,
    "N": 0.71,
    "O": 0.63,
    "F": 0.64,
    "Ne": 0.67,
    "Na": 1.55,
    "Mg": 1.39,
    "Al": 1.26,
    "Si": 1.16,
    "P": 1.11,
    "S": 1.03,
    "Cl": 0.99,
    "Ar": 0.96,
    "K": 1.96,
    "Ca": 1.71,
    "Sc": 1.48,
    "Ti": 1.36,
    "V": 1.34,
    "Cr": 1.22,
    "Mn": 1.19,
    "Fe": 1.16,
    "Co": 1.11,
    "Ni": 1.10,
    "Zn": 1.18,
    "Ga": 1.24,
    "Ge": 1.21,
    "As": 1.21,
    "Se": 1.16,
    "Br": 1.14,
    "Kr": 1.17,
    "Rb": 2.10,
    "Sr": 1.85,
    "Y": 1.63,
    "Zr": 1.54,
    "Nb": 1.47,
    "Mo": 1.38,
    "Tc": 1.28,
    "Ru": 1.25,
    "Rh": 1.25,
    "Pd": 1.20,
    "Ag": 1.28,
    "Cd": 1.36,
    "In": 1.42,
    "Sn": 1.40,
    "Sb": 1.40,
    "Te": 1.36,
    "I": 1.33,
    "Xe": 1.31,
}


def detect_linear(errortype, atom_types, cclib_data):
    """
    Check whether a linear molecule contain the right number of frequencies
    """

    linear_options_3 = [
        ["I", "I", "I"],
        ["N", "N", "N"],
        ["N", "C", "H"],
        ["O", "C", "O"],
        ["O", "C", "S"],
        ["S", "C", "S"],
        ["F", "Be", "F"],
        ["F", "Xe", "F"],
        ["O", "C", "N"],
        ["S", "C", "N"],
    ]
    linear_options_4 = [["C", "C", "H", "H"]]

    if len(atom_types) == 3:
        for linear_3 in linear_options_3:
            if (
                sorted(atom_types) == sorted(linear_3)
                and len(cclib_data["vibrations"]["frequencies"]) != 4
            ):
                errortype = "linear_mol_wrong"
                break
    elif len(atom_types) == 4:
        for linear_4 in linear_options_4:
            if (
                sorted(atom_types) == sorted(linear_4)
                and len(cclib_data["vibrations"]["frequencies"]) != 7
            ):
                errortype = "linear_mol_wrong"
                break

    return errortype


def full_check(w_dir_main=os.getcwd(), destination_fullcheck="", files="*.json", log=None):
    """
    Checks that multiple calculations were done following the same protocols, including
    program and version, grid size, level of theory, dispersion and solvation model.
    Parameters
    ----------
    w_dir_main : str
        Working directory
    destination_fullcheck : str
        Destination to create the file with the full check
    files : list of str
        json files to compare (glob.glob('*.json') and '*.json are both valid inputs to
        include all the json files from a folder)
    log : aqme.utils.Logger
        Logging instance where the status of the calculation will be written.
        If none provided it will default to aqme.utils.Logger('QCORR','fullcheck')
        and it will create the file QCORR_fullcheck.dat in the working directory.
    """

    if log is None: 
        log = Logger('QCORR','fullcheck')

    initial_dir = os.getcwd()
    w_dir_main = Path(w_dir_main)
    os.chdir(w_dir_main)

    if not isinstance(files, list):
        files = glob.glob(files)

    df_fullcheck = pd.DataFrame(
        columns=[
            "file",
            "program",
            "grid_type",
            "level_of_theory",
            "dispersion",
            "solvation",
        ]
    )

    for file in files:
        file_name = file.replace("/", "\\").split("\\")[-1].split(".")[0]
        with open(file) as json_file:
            cclib_data = json.load(json_file)

        program = cclib_data["metadata"]["QM program"]
        solvation = cclib_data["metadata"]["solvation"]
        dispersion = cclib_data["metadata"]["dispersion model"]
        grid_type = cclib_data["metadata"]["grid type"]
        functional = cclib_data["metadata"]["functional"]
        bs = cclib_data["metadata"]["basis set"]
        if functional != "" or bs != "":
            level_of_theory = "/".join([functional, bs])
        else:
            level_of_theory = ""
        # designed to detect G4 calcs
        if level_of_theory == "HF/GFHFB2":
            level_of_theory = "G4"
        df_fullcheck.loc[len(df_fullcheck.index)] = [
            file_name,
            program,
            grid_type,
            level_of_theory,
            dispersion,
            solvation,
        ]

    fullcheck_file = "--QCORR_Fullcheck_Analysis--.dat"
    fullcheck_txt = "\n-- Full check analysis --"
    for prop in df_fullcheck.columns:
        if prop != "file":
            unique_props = df_fullcheck[prop].unique()
            if len(unique_props) > 1:
                fullcheck_txt += f"\nx  Different {prop} used in the calculations:"
                for unique_prop in unique_props:
                    file_names = df_fullcheck["file"].loc[
                        df_fullcheck[prop] == unique_prop
                    ]
                    fullcheck_txt += f"\n     * {unique_prop} in:"
                    for file_name in file_names:
                        adapted_name = file_name.replace("/", "\\").split("\\")[-1]
                        fullcheck_txt += f"\n       - {adapted_name}"
            else:
                fullcheck_txt += (
                    f"\no  Same {prop} ({unique_props[0]}) used in all the calculations"
                )

    fullcheck_analysis = open(fullcheck_file, "w")
    fullcheck_analysis.write(fullcheck_txt)
    fullcheck_analysis.close()
    log.write(fullcheck_txt)

    if destination_fullcheck == "":
        destination_fullcheck = w_dir_main.joinpath("success/json_files/")
    else:
        destination_fullcheck = Path(destination_fullcheck)
    move_file(destination_fullcheck, w_dir_main, fullcheck_file)

    os.chdir(initial_dir)


def check_isomerization(isom_data, file):
    """
    Inputs two molecules with the atoms in the same order and checks if any bond
    is too different between them.

    Bonds are considered when the distance between two atoms is smaller than
    either the sum of their adjusted VDW radii or covalent radii (dist = n*R1 + n*R2, where
    n is a user defined parameter).

    Bonds forming part of TSs are removed.

    Parameters
    ----------
    isom_data : dict
            Contains data related to coordinates, atoms and connectivity for input and output files
    file : string
            Filename (with extension)

    Returns
    -------
    isomerized : bool
            True if there is a clearly distorted bond within the geometries
    """

    # load connectivity matrix from the starting points and convert string into matrix
    if not isom_data["Initial csv"].empty:
        filename = file.replace("_" + file.split("_")[-1], "")
        init_connectivity_string = isom_data["Initial csv"][
            isom_data["Initial csv"]["code_name"] == filename
        ]["initial_connectiv"][0]
        init_connectivity = json.loads(
            init_connectivity_string.replace(".", ",")
            .replace(",]", "],")
            .replace("],]", "]]")
        )
        isom_data["Atoms input"] = init_connectivity[0]

    else:
        init_connectivity = gen_connectivity(
            isom_data, isom_data["Atoms input"], isom_data["Coords input"]
        )

    # in case the systems are not the same
    if len(isom_data["Atoms output"]) != len(isom_data["Atoms input"]):
        isomerized = True
    else:
        final_connectivity = gen_connectivity(
            isom_data, isom_data["Atoms output"], isom_data["Coords output"]
        )

        # check connectivity differences from initial structure
        diff = final_connectivity - init_connectivity

        # remove bonds involved in TSs from connectivity matrixes
        if not isom_data["Initial csv"].empty:
            if "TS_atom_idx" in isom_data["Initial csv"].columns:
                ts_atoms = isom_data["Initial csv"][
                    isom_data["Initial csv"]["code_name"] == filename
                ]["TS_atom_idx"][0].split(",")
                for i, ts_idx in enumerate(ts_atoms):
                    for j, ts_idx_2 in enumerate(ts_atoms):
                        if j > i:
                            diff[int(ts_idx)][int(ts_idx_2)] = 0
                            diff[int(ts_idx_2)][int(ts_idx)] = 0

        isomerized = np.any(diff)

    return isomerized


def gen_connectivity(isom_data, atom_types_conn, COORDINATES_conn):
    """
    Use VDW radii to infer a connectivity matrix
    """

    conn_mat = np.zeros((len(atom_types_conn), len(atom_types_conn)))
    for i, elem_i in enumerate(atom_types_conn):
        for j, elem_j in enumerate(atom_types_conn):
            if j > i:
                vdw_ij = bondi[elem_i] + bondi[elem_j]
                rcov_ij = rcov[elem_i] + rcov[elem_j]
                dist_ij = np.linalg.norm(
                    np.array(COORDINATES_conn[i]) - np.array(COORDINATES_conn[j])
                )
                if dist_ij / vdw_ij < float(
                    isom_data["VdW radii fraction"]
                ) or dist_ij / rcov_ij < float(isom_data["Covalent radii fraction"]):
                    conn_mat[i][j] = 1
                else:
                    pass

    return conn_mat


def get_json_data(self, file, cclib_data):
    """
    Get metadata and GoodVibes data for the json file (for older versions of cclib)
    """

    outlines = read_file(os.getcwd(), self.args.w_dir_main, file)

    # initial loop just to detect the QM program
    for i, line in enumerate(outlines):
        # get program
        if line.strip() == "Cite this work as:":
            cclib_data["metadata"] = {}
            qm_program = outlines[i + 1]

            cclib_data["metadata"]["QM program"] = qm_program[1:-2]
            for j in range(i, i + 60):
                if "**********" in outlines[j]:
                    run_date = outlines[j + 2].strip()
                    cclib_data["metadata"]["run date"] = run_date
                    break
            break

        elif "* O   R   C   A *" in line:
            for j in range(i, i + 100):
                if "Program Version" in line.strip():
                    cclib_data["metadata"] = {}
                    version_program = "ORCA version " + line.split()[2]
                    cclib_data["metadata"]["QM program"] = version_program
                    break

    if cclib_data["metadata"]["QM program"].lower().find("gaussian") > -1:

        cclib_data["properties"]["rotational"] = {}
        for i, line in enumerate(outlines):
            # Extract memory
            if "%mem" in line:
                mem = line.strip().split("=")[-1]
                cclib_data["metadata"]["memory"] = mem

            # Extract number of processors
            elif "%nprocs" in line:
                nprocs = int(line.strip().split("=")[-1])
                cclib_data["metadata"]["processors"] = nprocs

            # Extract keywords line, solvation, dispersion and calculation type
            elif "#" in line and not hasattr(cclib_data, "keywords_line"):
                keywords_line = ""
                for j in range(i, i + 10):
                    if "----------" in outlines[j]:
                        break
                    else:
                        keywords_line += outlines[j].rstrip("\n")[1:]
                cclib_data["metadata"]["keywords line"] = keywords_line[2:]
                qm_solv, qm_disp = "gas_phase", "none"
                calc_type = "ground_state"
                calcfc_found, ts_found = False, False
                for keyword in keywords_line.split():
                    if keyword.lower().find("opt") > -1:
                        if keyword.lower().find("calcfc") > -1:
                            calcfc_found = True
                        if keyword.lower().find("ts") > -1:
                            ts_found = True
                    elif keyword.lower().startswith("scrf"):
                        qm_solv = keyword
                    elif keyword.lower().startswith("emp"):
                        qm_disp = keyword
                if calcfc_found and ts_found:
                    calc_type = "transition_state"
                cclib_data["metadata"]["solvation"] = qm_solv
                cclib_data["metadata"]["dispersion model"] = qm_disp
                cclib_data["metadata"]["ground or transition state"] = calc_type

            # Basis set name
            elif line[1:15] == "Standard basis":
                cclib_data["metadata"]["basis set"] = line.split()[2]

            # functional
            if not hasattr(cclib_data, "BOMD") and line[1:9] == "SCF Done":
                t1 = line.split()[2]
                if t1 == "E(RHF)":
                    cclib_data["metadata"]["functional"] = "HF"
                else:
                    cclib_data["metadata"]["functional"] = t1[
                        t1.index("(") + 2 : t1.rindex(")")
                    ]
                break

            # Extract grid type
            elif line[1:8] == "ExpMin=":
                grid_lookup = {
                    1: "sg1",
                    2: "coarse",
                    4: "fine",
                    5: "ultrafine",
                    7: "superfine",
                }
                IRadAn = int(line.strip().split()[-3])
                grid = grid_lookup[IRadAn]
                cclib_data["metadata"]["grid type"] = grid

        # Keeps track of convergence during Freq calcs
        if "optimization" in cclib_data:
            cclib_data["optimization"]["times converged"] = 1

        # Extract <S**2> before and after spin annihilation, energy, and convergence in freq calc
        for i in reversed(range(0, len(outlines) - 30)):
            # For time dependent (TD) calculations
            if "E(TD-HF/TD-DFT)" in outlines[i]:
                td_e = float(line.strip().split()[-1])
                cclib_data["properties"]["energy"][
                    "TD energy"
                ] = cclib.parser.utils.convertor(td_e, "hartree", "eV")

            # For G4 calculations look for G4 energies (Gaussian16a bug prints G4(0 K) as DE(HF)) --Brian modified to work for G16c-where bug is fixed.
            elif line.strip().startswith("E(ZPE)="):  # Overwrite DFT ZPE with G4 ZPE
                zero_point_corr = float(line.strip().split()[1])
            elif line.strip().startswith("G4(0 K)"):
                G4_energy = float(line.strip().split()[2])
                G4_energy -= zero_point_corr  # Remove G4 ZPE
                cclib_data["properties"]["energy"][
                    "G4 energy"
                ] = cclib.parser.utils.convertor(G4_energy, "hartree", "eV")

            # For ONIOM calculations use the extrapolated value rather than SCF value
            elif "ONIOM: extrapolated energy" in line.strip():
                oniom_e = float(line.strip().split()[4])
                cclib_data["properties"]["energy"][
                    "ONIOM energy"
                ] = cclib.parser.utils.convertor(oniom_e, "hartree", "eV")

            elif "S**2 before annihilation" in outlines[i]:
                cclib_data["properties"]["S2 after annihilation"] = float(
                    outlines[i].strip().split()[-1]
                )
                cclib_data["properties"]["S2 before annihilation"] = float(
                    outlines[i].strip().split()[-3][:-1]
                )

            # Extract symmetry point group
            elif "Full point group" in outlines[i]:
                point_group = outlines[i].strip().split()[3]
                cclib_data["properties"]["rotational"][
                    "symmetry point group"
                ] = point_group
                break

            elif "Stationary point found" in outlines[i]:
                cclib_data["optimization"]["times converged"] = 2

            # Extract symmetry number, rotational constants and rotational temperatures
            elif "Rotational symmetry number" in outlines[i]:
                symmno = int(outlines[i].strip().split()[3].split(".")[0])
                cclib_data["properties"]["rotational"]["symmetry number"] = symmno

            elif outlines[i].find("Rotational constants (GHZ):") > -1:
                try:
                    roconst = [
                        float(outlines[i].strip().replace(":", " ").split()[3]),
                        float(outlines[i].strip().replace(":", " ").split()[4]),
                        float(outlines[i].strip().replace(":", " ").split()[5]),
                    ]
                except ValueError:
                    if outlines[i].find("********") > -1:
                        roconst = [
                            float(outlines[i].strip().replace(":", " ").split()[4]),
                            float(outlines[i].strip().replace(":", " ").split()[5]),
                        ]
                cclib_data["properties"]["rotational"]["rotational constants"] = roconst

            elif outlines[i].find("Rotational temperature ") > -1:
                rotemp = [float(outlines[i].strip().split()[3])]
                cclib_data["properties"]["rotational"][
                    "rotational temperatures"
                ] = rotemp

            elif outlines[i].find("Rotational temperatures") > -1:
                try:
                    rotemp = [
                        float(outlines[i].strip().split()[3]),
                        float(outlines[i].strip().split()[4]),
                        float(outlines[i].strip().split()[5]),
                    ]
                except ValueError:
                    if outlines[i].find("********") > -1:
                        rotemp = [
                            float(outlines[i].strip().split()[4]),
                            float(outlines[i].strip().split()[5]),
                        ]
                cclib_data["properties"]["rotational"][
                    "rotational temperatures"
                ] = rotemp

            elif outlines[i].find("SCF GIAO Magnetic shielding tensor (ppm)") > -1:
                nmr_iso = []
                nmr_anis = []
                nmr_eigen = []
                cclib_data["properties"]["NMR"] = {}
                for j in range(i, len(outlines)):
                    if outlines[j].find("Isotropic") > -1:
                        nmr_iso.append(float(outlines[j].split()[4]))
                        nmr_anis.append(float(outlines[j].split()[7]))
                    elif outlines[j].find("Eigenvalues") > -1:
                        nmr_eigen.append(
                            [
                                float(outlines[j].split()[1]),
                                float(outlines[j].split()[2]),
                                float(outlines[j].split()[3]),
                            ]
                        )
                    elif outlines[j].find("*************************") > -1:
                        break
                cclib_data["properties"]["NMR"]["NMR anisotopic tensors"] = nmr_anis
                cclib_data["properties"]["NMR"]["NMR eigenvalues"] = nmr_eigen
                cclib_data["properties"]["NMR"]["NMR isotopic tensors"] = nmr_iso


    elif cclib_data["metadata"]["QM program"].lower().find("orca") > -1:
        for i in reversed(range(0, outlines)):
            if outlines[i][:25] == "FINAL SINGLE POINT ENERGY":
                # in eV to match the format from cclib
                orca_e = float(outlines[i].split()[-1])
                cclib_data["properties"]["energy"][
                    "final single point energy"
                ] = cclib.parser.utils.convertor(orca_e, "hartree", "eV")
                break

    if cclib_data != {}:
        with open(f'{file.split(".")[0]}.json', "w") as outfile:
            json.dump(cclib_data, outfile, indent=1)

    return cclib_data



#####################################################.
#        This file stores all the functions         #
#            used in CSEARCH-FullMonte              #
#####################################################.

import sys
import numpy as np
import math
import random
from rdkit.Chem import AllChem as Chem
from rdkit.Chem import rdMolTransforms, rdMolAlign

from aqme.utils import set_metal_atomic_number, get_conf_RMS
from aqme.csearch.utils import minimize_rdkit_energy


def realign_mol(
    mol, conf, coord_Map, alg_Map, mol_template, maxsteps
):  # RAUL: This function requires a clear separation between minimization and alignment.
    """
    Minimizes and aligns the molecule provided freezing the atoms that match the mol_template

    Parameters
    ----------
    mol : RDKit mol object
        Molecule to be minimized and aligned
    conf : int
        Number that indicates which conformation of the molecule will be minimized and aligned
    coord_Map : [type]
        [description]
    alg_Map : [type]
        [description]
    mol_template : [type]
        [description]
    maxsteps : int
        Maximum number of iterations in FF minimization

    Returns
    -------
    mol,energy
        The updated mol object and the final forcefield energy.
    """

    num_atom_match = mol.GetSubstructMatch(mol_template)
    forcefield = Chem.UFFGetMoleculeForceField(mol, confId=conf)
    for i, idxI in enumerate(num_atom_match):
        for idxJ in num_atom_match[i + 1 :]:
            d = coord_Map[idxI].Distance(coord_Map[idxJ])
            forcefield.AddDistanceConstraint(idxI, idxJ, d, d, 10000)
    forcefield.Initialize()
    forcefield.Minimize(maxIts=maxsteps)
    # rotate the embedded conformation onto the core_mol:
    rdMolAlign.AlignMol(
        mol,
        mol_template,
        prbCid=conf,
        refCid=-1,
        atomMap=alg_Map,
        reflect=True,
        maxIters=100,
    )
    energy = float(forcefield.CalcEnergy())
    return mol, energy


def rotate_dihedrals(conformer, dihedrals, seed, stepsize):
    """
    Applies a random rotation to all the dihedrals

    Parameters
    ----------
    conformer : rdkit.Chem.rdchem.Conformer
        The conformer whose angles are going to be rotated (conformer = mol.GetConformer(cid))
    dihedrals : list
        A list of tuples of all the dihedrals that are going to be rotated.
    seed : int
        seed for the random module
    stepsize : float
        Angle in Degrees to do the steps between 0.0 and 360.0
    """

    rad_range = np.arange(stepsize, 360.0, stepsize)
    for dihedral in dihedrals:
        random.seed(seed)  # RAUL: Any good reason to keep reseting the seed?
        rad_ang = random.choice(rad_range)
        rad = math.pi * rad_ang / 180.0
        rdMolTransforms.SetDihedralRad(conformer, *dihedral, value=rad)


def generating_conformations_fullmonte(
    name,
    args,
    rotmatches,
    selectedcids_rdkit,
    outmols,
    sdwriter,
    dup_data,
    dup_data_idx,
    coord_Map,
    alg_Map,
    mol_template,
    ff,
):

    ##working with fullmonte
    n_unique_conformers = len(selectedcids_rdkit)
    args.log.write(f"\no  Generation of confomers using FULLMONTE using {n_unique_conformers} unique conformer(s) as starting point(s)")

    # Writing the conformers as mol objects to sdf
    sdtemp = Chem.SDWriter(name + "_" + "rdkit" + args.output)
    for conf in selectedcids_rdkit:
        sdtemp.write(outmols[conf], conf)
    sdtemp.close()

    fmmols = Chem.SDMolSupplier(name + "_" + "rdkit" + args.output, removeHs=False)
    if fmmols is None:
        args.log.write("Could not open " + name + args.output)
        args.log.finalize()
        sys.exit()

    # array for each each unique from rdkit
    unique_mol, c_energy, unique_mol_sample = [], [], []

    # STEP 1: Use start conformation for and append to unique list
    nsteps = 1
    for mol_fm in fmmols:
        unique_mol.append(mol_fm)
        c_energy.append(float(mol_fm.GetProp("Energy")))

    # defining unique mol sample for choosing
    globmin = min(c_energy)
    for ene in reversed(c_energy):
        if abs(globmin - ene) < args.ewin_sample_fullmonte:
            unique_mol_sample.append(unique_mol[c_energy.index(ene)])

    while nsteps < args.nsteps_fullmonte + 1:
        seed = nsteps

        # STEP 2: Choose mol object form unique_mol:
        random.seed(seed)
        mol_rot = random.choices(unique_mol_sample, k=1)[0]

        # updating the location of mol object i.e., the hexadecimal locaiton to a new one so the older one isnt affected
        mol = Chem.RWMol(mol_rot)
        rot_mol = mol.GetMol()

        # STEP 3: Choose random subset of dihedral from rotmatches
        random.seed(seed)  # RAUL: Any good reason to keep reseting the seed ?
        k = min(len(rotmatches), args.nrot_fullmonte)
        mutable_dihedrals = random.choices(rotmatches, k=k)

        # STEP 4: for the given conformation, then apply a random rotation to each torsion in the subset
        conformer = rot_mol.GetConformer()
        rotate_dihedrals(conformer, mutable_dihedrals, seed, args.ang_fullmonte)

        # STEP 5: Optimize geometry rot_mol
        if (coord_Map, alg_Map, mol_template) == (None, None, None):
            energy = minimize_rdkit_energy(
                rot_mol, -1, args.log, ff, args.opt_steps_rdkit
            )
        else:
            mol, energy = realign_mol(
                rot_mol, -1, coord_Map, alg_Map, mol_template, args.opt_steps_rdkit
            )

        # STEP 6 : Check for DUPLICATES - energy and rms filter (reuse)
        #  if the conformer is unique then save it the list
        exclude_conf = False
        # compare against allprevious conformers located
        for j, seenmol in enumerate(unique_mol):
            if abs(energy - c_energy[j]) < args.initial_energy_threshold:
                exclude_conf = True
                break
            if abs(energy - c_energy[j]) < args.energy_threshold:
                rms = get_conf_RMS(
                    rot_mol, seenmol, -1, -1, args.heavyonly, args.max_matches_rmsd
                )
                if rms < args.rms_threshold:
                    exclude_conf = True
                    break
        if not exclude_conf:
            unique_mol.append(rot_mol)
            c_energy.append(energy)
            unique_mol[c_energy.index(energy)].SetProp("Energy", str(energy))

        unique_mol_sample = []
        # STEP 7: ANALYSE THE UNIQUE list for lowest energy, reorder the uniques if greater the given thershold remove
        globmin = min(c_energy)
        for ene in reversed(c_energy):
            indx = c_energy.index(ene)
            if abs(globmin - ene) > args.ewin_fullmonte:
                unique_mol.pop(indx)
                c_energy.pop(indx)
            if abs(globmin - ene) < args.ewin_sample_fullmonte:
                unique_mol_sample.append(unique_mol[indx])

        nsteps += 1

    dup_data.at[dup_data_idx, "FullMonte-Unique-conformers"] = len(unique_mol)

    cids = list(range(len(unique_mol)))
    sorted_all_cids = sorted(cids, key=lambda cid: c_energy[cid])

    # STEP 9: WRITE FINAL uniques to sdf for xtb or ani
    for i, cid in enumerate(sorted_all_cids):
        unique_mol[cid].SetProp("_Name", name + " " + str(i))
        if coord_Map is None and alg_Map is None and mol_template is None:
            if len(args.metal_atoms) >= 1:
                set_metal_atomic_number(unique_mol[cid], args.metal_idx, args.metal_sym)
            sdwriter.write(unique_mol[cid])
        else:
            mol_realigned, _ = realign_mol(
                unique_mol[cid],
                -1,
                coord_Map,
                alg_Map,
                mol_template,
                args.opt_steps_rdkit,
            )
            if len(args.metal_atoms) >= 1:
                set_metal_atomic_number(mol_realigned, args.metal_idx, args.metal_sym)
            sdwriter.write(mol_realigned)

    status = 1

    return status



"""
Parameters
----------

   files : list of str, default=''
      Filenames of QM output files to analyze. If *.log (or other strings that 
      are not lists such as *.out) are specified, the program will look for all 
      the log files in the working directory through glob.glob(*.log)
   w_dir_main : str, default=os.getcwd()
      Working directory
   fullcheck : bool, default=True
      Perform an analysis to detect whether the calculations were done 
      homogeneously (i.e. same level of theory, solvent, grid size, etc)
   varfile : str, default=None
      Option to parse the variables using a yaml file (specify the filename)
   ifreq_cutoff : float, default=0.0
      Cut off for to consider whether a frequency is imaginary (absolute of the 
      specified value is used)
   amplitude_ifreq : float, default=0.2
      Amplitude used to displace the imaginary frequencies to fix
   freq_conv : str, default=None
      If a string is defined, it will remove calculations that converged during 
      optimization but did not convergence in the subsequent frequency 
      calculation. Options: opt keyword as string (i.e. 'opt=(calcfc,maxstep=5)'). 
      If readfc is specified in the string, the chk option must be included as well.
   s2_threshold : float, default=10.0
      Cut off for spin contamination during analysis in % of the expected value 
      (i.e. multiplicity 3 has an the expected <S**2> of 2.0, 
      if s2_threshold = 10, the <S**2> value is allowed to be 2.0 +- 0.2). 
      Set s2_threshold = 0 to deactivate this option.
   dup_threshold : float, default=0.0001
      Energy (in hartree) used as the energy difference in E, H and G to detect 
      duplicates
   isom_type : str, default=None
      Check for isomerization from the initial input file to the resulting 
      output files. It requires the extension of the initial input files 
      (i.e. isom_type='com' or 'gjf') and the folder of the input files must be 
      added in the isom_inputs option
   isom_inputs : str, default=os.getcwd()
      Folder containing the initial input files to check for isomerization
   vdwfrac : float, default=0.50
      Fraction of the summed VDW radii that constitutes a bond between two atoms 
      in the isomerization filter
   covfrac : float, default=1.10
      Fraction of the summed covalent radii that constitutes a bond between two 
      atoms in the isomerization filter
      
.. note::

   New input files are generated through the QPREP module and, therefore, all 
   QPREP arguments can be used when calling QCORR and will overwrite default 
   options. For example, if the user specifies qm_input='wb97xd/def2svp', 
   all the new input files generated to fix issues will contain this keywords 
   line. See examples in the 'Example_workflows' folder for more information.

"""
######################################################.
#        This file stores the QCORR class            #
######################################################.

import os
import sys
import glob
import time
import pandas as pd
import json
import subprocess
import numpy as np

try:
    import cclib
except ModuleNotFoundError:
    print("x  cclib is not installed! You can install the program with 'conda install -c conda-forge cclib' or 'pip install cclib'")
    sys.exit()
from pathlib import Path
from aqme.utils import (
    move_file,
    QM_coords,
    get_info_input,
    load_variables,
    read_file,
    cclib_atoms_coords,
)
from aqme.qcorr_utils import (
    detect_linear,
    check_isomerization,
    full_check,
    get_json_data,
)
from aqme.qprep import qprep


class qcorr:
    """
    Class containing all the functions from the QCORR module.

    Parameters
    ----------
    kwargs : argument class
        Specify any arguments from the QCORR module (for a complete list of variables, visit the AQME documentation)
    """

    def __init__(self, **kwargs):

        # load default and user-specified variables
        self.args = load_variables(kwargs, "qcorr")

        if len(self.args.files) == 0:
            self.args.log.write('\nx  No files were found! Make sure you use quotation marks if you are using * (i.e. --files "*.log")')
            self.args.log.finalize()
            sys.exit()

        # QCORR analysis
        if self.args.files[0].split('.')[1].lower() not in ['log','out','json']:
            self.args.log.write(f"\nx  The format used ({self.args.files[0].split('.')[1].lower()}) is not compatible with QCORR! Formats accepted: log, out, json")
            self.args.log.finalize()
            sys.exit()

        self.qcorr_processing()

        # this is added to avoid path problems in jupyter notebooks
        os.chdir(self.args.initial_dir)

    def qcorr_processing(self):
        """
        General function of the QCORR module that:

        1. Analyzes the QM output files and moves output files with normal termination and no extra imaginary frequencies to the same folder
        2. Generates input files to fix errors and extra imaginary frequencies
        3. Generates input files with new keywords line(s) from the normally terminated files from point 1 (i.e. single-point energy corrections)
        """

        start_time_overall = time.time()

        # generate some data
        file_terms = {
            "finished": 0,
            "sp_calcs": 0,
            "extra_imag_freq": 0,
            "ts_no_imag_freq": 0,
            "freq_no_conv": 0,
            "spin_contaminated": 0,
            "duplicate_calc": 0,
            "atom_error": 0,
            "scf_error": 0,
            "no_data": 0,
            "linear_mol_wrong": 0,
            "not_specified": 0,
            "geom_rules_qcorr": 0,
            "isomerized": 0,
        }

        duplicate_data = {
            "File": [],
            "Energies": [],
            "Enthalpies": [],
            "Gibbs": [],
            "RO_constant": [],
        }

        self.args.log.write(f"o  Analyzing output files in {self.args.w_dir_main}\n")
        os.chdir(self.args.w_dir_main)
        # analyze files
        for file in sorted(self.args.files):
            # get initial cclib data and termination/error types and discard calcs with no data
            file_name = os.path.basename(file).split(".")[0]
            termination, errortype, cclib_data, outlines, file = self.cclib_init(
                file, file_name
            )
            if errortype in ["no_data", "atomicbasiserror"]:
                file_terms, _ = self.organize_outputs(
                    file, termination, errortype, file_terms
                )
                if errortype == "atomicbasiserror":
                    os.remove(file_name + ".json")
                    self.args.log.write(f"{os.path.basename(file)}: Termination = {termination}, Error type = {errortype}")
                continue

            # check for duplicates and fix wrong number of freqs in normally terminated calculations and
            elif termination == "normal":
                (
                    atom_types,
                    cartesians,
                    duplicate_data,
                    errortype,
                    cclib_data,
                    dup_off,
                ) = self.analyze_normal(
                    duplicate_data, errortype, cclib_data, file_name
                )

            # fix calcs that did not terminated normally

            elif termination != "normal":
                atom_types, cartesians, cclib_data = self.analyze_abnormal(
                    errortype, cclib_data, outlines
                )

            # check for isomerization
            if self.args.isom_type is not None:
                errortype = self.analyze_isom(file, cartesians, atom_types, errortype)

            # move initial QM input files (if the files are placed in the same folder as the output files)
            if (
                os.path.exists(f"{self.args.w_dir_main}/{file_name}.com")
                and self.args.round_num == 1
            ):
                move_file(
                    self.args.w_dir_main.joinpath("inputs/"),
                    self.args.w_dir_main,
                    f"{file_name}.com",
                )

            # create input files through QPREP to fix the errors (some errors require user intervention)
            if errortype not in [
                "ts_no_imag_freq",
                "isomerization",
                "duplicate_calc",
                "spin_contaminated",
                "none",
                "sp_calc",
            ]:
                self.qcorr_fixing(cclib_data, file, atom_types, cartesians)

            # This part places the calculations and json files in different folders depending on the type of termination
            if errortype == "duplicate_calc":
                self.args.log.write(f"{os.path.basename(file)}: Termination = {termination}, Error type = {errortype}, Duplicate of = {dup_off}")
            else:
                self.args.log.write(f"{os.path.basename(file)}: Termination = {termination}, Error type = {errortype}")

            file_terms, destination = self.organize_outputs(
                file, termination, errortype, file_terms
            )

            if errortype in ["none", "sp_calc"]:
                destination_json = destination.joinpath("json_files/")
                move_file(destination_json, self.args.w_dir_main, file_name + ".json")
            else:
                os.remove(file_name + ".json")

            # write information about the QCORR analysis in a csv
            csv_qcorr = self.write_qcorr_csv(file_terms)

        # performs a full analysis to ensure that the calcs were run with the same parameters
        if self.args.fullcheck == "False":
            self.args.fullcheck = False
        elif self.args.fullcheck == "True":
            self.args.fullcheck = True
        if self.args.fullcheck:
            no_normal_terms = False
            try:
                df_qcorr = pd.read_csv(csv_qcorr)
                if df_qcorr["Normal termination"][0] > 0:
                    json_files = glob.glob(f"{destination_json}/*.json")
                    full_check(
                        w_dir_main=destination_json,
                        destination_fullcheck=destination_json,
                        files=json_files,
                        log=self.args.log,
                    )
                else:
                    no_normal_terms = True
            except UnboundLocalError:
                no_normal_terms = True
            if no_normal_terms:
                self.args.log.write("\nx  No normal terminations with no errors to run the full check analysis")

        elapsed_time = round(time.time() - start_time_overall, 2)
        self.args.log.write(f"\n Time QCORR: {elapsed_time} seconds\n")
        self.args.log.finalize()

        # NOT needed as already created in initial_dir
        # move dat and csv file containing the QCORR information if this is a sequential QCORR analysis
        # if self.args.resume_qcorr:
        #     destination_data = self.args.w_dir_main.joinpath("../../../")
        #     move_file(
        #         destination_data,
        #         self.args.w_dir_main,
        #         f"QCORR-run_{self.args.round_num}.dat",
        #     )
        #     move_file(
        #         destination_data,
        #         self.args.w_dir_main,
        #         f"QCORR-run_{self.args.round_num}-stats.csv",
        #     )

    # include geom filters (ongoing work)

    # 			if len(self.args.geom_rules) >= 1:
    # 				passing_rules = True
    # 				valid_mol_gen = True
    # 				self.args.log.write("  ----- geom_rules filter(s) will be applied to the output file -----\n")
    # 				try:
    # 					format_file = file.split('.')[1]
    # 					mol = output_to_mol(file,format_file)
    # 					print_error_geom_rules=False
    # 					if ob_compat and rdkit_compat:
    # 						passing_rules = geom_rules_output(mol,self.args,self.args.log,file,print_error_geom_rules)
    # 						if not passing_rules:
    # 							errortype = 'fail_geom_rules'
    # 					os.remove(file.split('.')[0]+'.mol')
    # 				except AttributeError:
    # 					valid_mol_gen = False
    # 					os.remove(file.split('.')[0]+'.mol')
    # 					self.args.log.write("The file could not be converted into a mol object, geom_rules filter(s) will be disabled\n")

    def cclib_init(self, file, file_name):
        """
        Determine termination and error types (initial determination), create json files
        with cclib and load the data in the cclib json files
        """

        # cclib generation of json files with ccwrite
        termination, errortype, cclib_data, file = self.json_gen(file, file_name)
        outlines = []

        if errortype == "no_data":
            return termination, errortype, None, None, file 

        # calculations with 1 atom
        if cclib_data["properties"]["number of atoms"] == 1:
            cclib_data["vibrations"] = {"frequencies": [], "displacement": []}
            if not "energy" in cclib_data["properties"]:
                termination = "other"
                errortype = "not_specified"
            elif not "free energy" in cclib_data["properties"]["energy"]:
                errortype = "sp_calc"

        # general errors
        elif "vibrations" not in cclib_data:
            termination = "other"
            errortype = "not_specified"
            if "optimization" in cclib_data:
                # if the optimization finished, only a freq job is required
                if (
                    "done" in cclib_data["optimization"]
                    and cclib_data["optimization"]["done"]
                ):
                    errortype = "no_freq"

            # use very short reversed loop to find basis set incompatibilities and SCF errors
            outlines = read_file(os.getcwd(), self.args.w_dir_main, file)
            for i in reversed(range(len(outlines) - 15, len(outlines))):
                if (
                    outlines[i].find("Normal termination") > -1
                    and errortype != "no_freq"
                ):
                    termination = "normal"
                    errortype = "sp_calc"
                    cclib_data["metadata"][
                        "ground or transition state"
                    ] = "SP calculation"
                    break
                elif (
                    outlines[i - 1].find("Atomic number out of range") > -1
                    or outlines[i - 1].find("basis sets are only available") > -1
                ):
                    errortype = "atomicbasiserror"
                    break
                elif outlines[i].find("SCF Error") > -1:
                    errortype = "SCFerror"
                    break

        # normal terminations
        if "vibrations" in cclib_data or errortype == "sp_calc":
            # spin contamination analysis using user-defined thresholds
            if "S2 after annihilation" in cclib_data["properties"]:
                unpaired_e = cclib_data["properties"]["multiplicity"] - 1
                # this first part accounts for singlet diradicals (threshold is 10% of the spin before annihilation)
                if unpaired_e == 0:
                    if (
                        float(cclib_data["properties"]["S2 after annihilation"])
                        > abs(float(self.args.s2_threshold) / 100)
                        * cclib_data["properties"]["S2 before annihilation"]
                    ):
                        errortype = "spin_contaminated"
                else:
                    spin = unpaired_e * 0.5
                    s2_expected_value = spin * (spin + 1)
                    spin_diff = abs(
                        float(cclib_data["properties"]["S2 after annihilation"])
                        - s2_expected_value
                    )
                    if (
                        spin_diff
                        > abs(float(self.args.s2_threshold) / 100) * s2_expected_value
                    ):
                        errortype = "spin_contaminated"

        return termination, errortype, cclib_data, outlines, file 

    def analyze_normal(self, duplicate_data, errortype, cclib_data, file_name):
        """
        Analyze errors from normally terminated calculations
        """
        atom_types, cartesians = cclib_atoms_coords(cclib_data)
        dup_off = None
        if errortype == "none":
            # in eV, converted to hartree using the conversion factor from cclib
            E_dup = cclib_data["properties"]["energy"]["total"]
            E_dup = cclib.parser.utils.convertor(E_dup, "eV", "hartree")
            # in hartree
            try:
                H_dup = cclib_data["properties"]["enthalpy"]
                G_dup = cclib_data["properties"]["energy"]["free energy"]
            except (AttributeError, KeyError):
                if cclib_data["properties"]["number of atoms"] == 1:
                    if cclib_data["metadata"]["keywords line"].find("freq") == -1:
                        errortype = "sp_calc"
                        cclib_data["metadata"][
                            "ground or transition state"
                        ] = "SP calculation"
                    H_dup = E_dup
                    G_dup = E_dup
            try:
                ro_dup = cclib_data["properties"]["rotational"]["rotational constants"]
                if len(ro_dup) != 3:
                    ro_dup = None
            except:
                ro_dup = None

            # detects if this calculation is a duplicate
            for i, _ in enumerate(duplicate_data["Energies"]):
                E_diff = abs(E_dup - duplicate_data["Energies"][i])
                H_diff = abs(H_dup - duplicate_data["Enthalpies"][i])
                G_diff = abs(G_dup - duplicate_data["Gibbs"][i])
                if (ro_dup is not None) and (
                    duplicate_data["RO_constant"][i] is not None
                ):
                    ro_diff = np.linalg.norm(
                        np.array(ro_dup) - np.array(duplicate_data["RO_constant"][i])
                    )
                if max([E_diff, H_diff, G_diff]) < abs(float(self.args.dup_threshold)):
                    if (ro_dup is not None) and (ro_diff < self.args.ro_threshold):
                        errortype = "duplicate_calc"
                        dup_off = duplicate_data["File"][i]

        if errortype == "none":
            duplicate_data["File"].append(file_name)
            duplicate_data["Energies"].append(E_dup)
            duplicate_data["Enthalpies"].append(H_dup)
            duplicate_data["Gibbs"].append(G_dup)
            duplicate_data["RO_constant"].append(ro_dup)

            initial_ifreqs = 0
            for freq in cclib_data["vibrations"]["frequencies"]:
                if float(freq) < 0 and abs(float(freq)) > abs(
                    float(self.args.ifreq_cutoff)
                ):
                    initial_ifreqs += 1

            # exclude TS imag frequency
            if (
                cclib_data["metadata"]["ground or transition state"]
                == "transition_state"
            ):
                initial_ifreqs -= 1

            # gives new coordinates by displacing the normal mode(s) of the negative freq(s)
            if initial_ifreqs > 0:
                errortype = "extra_imag_freq"

            elif initial_ifreqs < 0:
                errortype = "ts_no_imag_freq"

            if len(atom_types) in [3, 4]:
                errortype = detect_linear(errortype, atom_types, cclib_data)

            # detects no convergence issues during freq calcs
            if self.args.freq_conv is not None:
                if (
                    errortype == "none"
                    and cclib_data["optimization"]["times converged"] == 1
                ):
                    errortype = "freq_no_conv"

        if errortype in ["extra_imag_freq", "freq_no_conv", "linear_mol_wrong"]:
            if errortype == "extra_imag_freq":
                cartesians = self.fix_imag_freqs(cclib_data, cartesians)

            # in case no previous OPT was done (only works if it's not a TS)
            opt_found = False
            for keyword in cclib_data["metadata"]["keywords line"].split():
                if keyword.lower().startswith("opt"):
                    opt_found = True

            if not opt_found:
                cclib_data["metadata"]["keywords line"] += " opt"

            if errortype == "freq_no_conv":
                # adjust the keywords so only FREQ is calculated
                new_keywords_line = ""
                for keyword in cclib_data["metadata"]["keywords line"].split():
                    if keyword.lower().startswith("opt"):
                        keyword = self.args.freq_conv
                        if (
                            cclib_data["metadata"]["ground or transition state"]
                            == "transition_state"
                        ):
                            keyword = keyword.replace("=(", "=(ts,noeigen,")
                    new_keywords_line += keyword
                    new_keywords_line += " "
                cclib_data["metadata"]["keywords line"] = new_keywords_line

            elif errortype == "linear_mol_wrong":
                cclib_data["metadata"]["keywords line"] += " symmetry=(PG=Cinfv)"

        return atom_types, cartesians, duplicate_data, errortype, cclib_data, dup_off

    def analyze_abnormal(self, errortype, cclib_data, outlines):
        """
        Analyze errors from calculations that did not finish normally
        """
        # for calcs with finished OPT but no freqs, adjust the keywords so only FREQ is calculated
        if errortype == "no_freq":
            new_keywords_line = ""
            for keyword in cclib_data["metadata"]["keywords line"].split():
                if keyword.lower().startswith("opt"):
                    keyword = ""
                else:
                    new_keywords_line += keyword
                    new_keywords_line += " "
            cclib_data["metadata"]["keywords line"] = new_keywords_line
            atom_types, cartesians = cclib_atoms_coords(cclib_data)
        else:
            # help to fix SCF convergence errors
            if errortype == "SCFerror":
                if (
                    cclib_data["metadata"]["keywords line"].find(" scf=xqc") > -1
                    or cclib_data["metadata"]["keywords line"].find(" scf=qc") > -1
                ):
                    new_keywords_line = ""
                    for keyword in cclib_data["metadata"]["keywords line"].split():
                        if keyword == "scf=xqc":
                            keyword = "scf=qc"
                        new_keywords_line += keyword
                        new_keywords_line += " "
                    cclib_data["metadata"]["keywords line"] = new_keywords_line

                else:
                    cclib_data["metadata"]["keywords line"] += " scf=xqc"

            if errortype in ["not_specified", "SCFerror"]:
                if "geometric values" in cclib_data["optimization"]:
                    RMS_forces = [
                        row[1] for row in cclib_data["optimization"]["geometric values"]
                    ]
                    # cclib uses None when the values are corrupted in the output files, replace None for a large number
                    RMS_forces = [10000 if val is None else val for val in RMS_forces]
                    min_RMS = RMS_forces.index(min(RMS_forces))
                else:
                    # for optimizations that fail in the first step
                    min_RMS = 0

                atom_types, cartesians = QM_coords(
                    outlines,
                    min_RMS,
                    cclib_data["properties"]["number of atoms"],
                    "gaussian",
                    cclib_data["metadata"]["keywords line"],
                )

        return atom_types, cartesians, cclib_data

    def analyze_isom(self, file, cartesians, atom_types, errortype):
        """
        Check if the initial structure isomerized during QM geometry optimization
        """

        isomerized = False
        isom_valid = True
        init_csv = pd.DataFrame()
        try:
            os.chdir(self.args.isom_inputs)
        except FileNotFoundError:
            self.args.log.write("x  The PATH specified in isom_inputs doesn't exist!")
            isom_valid = False

        if not isom_valid:
            os.chdir(self.args.initial_dir)
            self.args.log.finalize()
            sys.exit()

        try:
            atoms_com, coords_com, atoms_and_coords = [], [], []
            if len(self.args.isom_type.split(".")) == 1:
                atoms_and_coords, _, _ = get_info_input(
                    f'{file.split(".")[0]}.{self.args.isom_type}'
                )

            elif self.args.isom_type.split(".")[1] != "csv":
                init_csv = pd.read_csv(self.args.isom_type)

            for line in atoms_and_coords:
                atoms_com.append(line.split()[0])
                coords_com.append(
                    [
                        float(line.split()[1]),
                        float(line.split()[2]),
                        float(line.split()[3]),
                    ]
                )

            isom_data = {
                "Coords input": coords_com,
                "Coords output": cartesians,
                "Atoms input": atoms_com,
                "Atoms output": atom_types,
                "VdW radii fraction": self.args.vdwfrac,
                "Covalent radii fraction": self.args.covfrac,
                "Initial csv": init_csv,
            }

            isomerized = check_isomerization(isom_data, file)

        except FileNotFoundError:
            self.args.log.write(f"x  No com file were found for {os.path.basename(file)}, the check_geom test will be disabled for this calculation")

        if isomerized:
            errortype = "isomerization"

        os.chdir(self.args.w_dir_main)

        return errortype

    def qcorr_fixing(self, cclib_data, file, atom_types, cartesians):
        """
        Create com files for resubmission with the suggested protocols to correct the errors
        """

        # user-defined keywords line, mem and nprocs overwrites previously used parameters
        if self.args.qm_input != "":
            cclib_data["metadata"]["keywords line"] = self.args.qm_input

        if self.args.mem != "16GB":
            cclib_data["metadata"]["memory"] = self.args.mem
        elif "memory" not in cclib_data["metadata"]:
            cclib_data["metadata"]["memory"] = "16GB"

        if self.args.nprocs != 8:
            cclib_data["metadata"]["processors"] = self.args.nprocs
        elif "processors" not in cclib_data["metadata"]:
            cclib_data["metadata"]["processors"] = 8

        if self.args.resume_qcorr:
            destination_fix = Path(
                f"{self.args.w_dir_main}/../../run_{self.args.round_num}/fixed_QM_inputs"
            )
        else:
            destination_fix = Path(
                f"{self.args.w_dir_main}/failed/run_{self.args.round_num}/fixed_QM_inputs"
            )

        if cclib_data["metadata"]["QM program"].lower().find("gaussian") > -1:
            program = "gaussian"
        elif cclib_data["metadata"]["QM program"].lower().find("orca") > -1:
            program = "orca"

        if program in ["gaussian", "orca"]:
            qprep(
                destination=destination_fix,
                w_dir_main=self.args.w_dir_main,
                files=os.path.basename(file),
                charge=cclib_data["properties"]["charge"],
                mult=cclib_data["properties"]["multiplicity"],
                program=program,
                atom_types=atom_types,
                cartesians=cartesians,
                qm_input=cclib_data["metadata"]["keywords line"],
                mem=cclib_data["metadata"]["memory"],
                nprocs=cclib_data["metadata"]["processors"],
                chk=self.args.chk,
                qm_end=self.args.qm_end,
                bs_gen=self.args.bs_gen,
                bs_nogen=self.args.bs_nogen,
                gen_atoms=self.args.gen_atoms,
                create_dat=False,
            )
        else:
            self.args.log.write(f"x  Couldn't create an input file to fix {os.path.basename(file)} (compatible programs: Gaussian and ORCA)\n")

    def json_gen(self, file, file_name):
        """
        Create a json file with cclib and load a dictionary
        """

        termination, errortype = "normal", "none"

        command_run_1 = ["ccwrite", "json", file]
        subprocess.run(command_run_1, capture_output=True)

        cclib_data = {}
        try:
            with open(file_name + ".json") as json_file:
                cclib_data = json.load(json_file)
        except FileNotFoundError:
            try:
                # this part avoids problems when using cclib from command lines (not complete file PATH)
                file = f'{self.args.initial_dir}/{file}'
                command_run_2 = ["ccwrite", "json", file]
                subprocess.run(command_run_2, capture_output=True)
                with open(file_name + ".json") as json_file:
                    cclib_data = json.load(json_file)
            except FileNotFoundError:
                termination = "other"
                errortype = "no_data"

        # add parameters that might be missing from cclib (depends on the version)
        if not hasattr(cclib_data, "metadata") and errortype != "no_data":
            cclib_data = get_json_data(self, file, cclib_data)

        # this is just a "dirty hack" until cclib is updated to be compatible for print mini in ORCA
        if hasattr(cclib_data, "metadata"):
            if cclib_data["metadata"]["QM program"].lower().find("orca") > -1:
                if "final single point energy" in cclib_data["properties"]["energy"]:
                    termination, errortype = "normal", "none"

        if errortype == "no_data":
            self.args.log.write(f"x  Potential cclib compatibility problem or no data found for file {file_name} (Termination = {termination}, Error type = {errortype})")

        return termination, errortype, cclib_data, file

    def fix_imag_freqs(self, cclib_data, cartesians):
        """
        Fixes undersired (extra) imaginary frequencies from QM calculations. This function multiplies the imaginary normal mode vectors by the selected amplitude (0.2 is the default amplitude in the pyQRC script from GitHub, user: bobbypaton).	By default, all the extra imaginary modes are used (i.e. in calculations with three	extra imaginary frequencies, all the three modes will be used to displace the atoms). This can be tuned with the --ifreq_cutoff option (i.e. only use freqs lower than -50 cm-1).

        Parameters
        ----------
        cclib_data : cclib object
            Variables parsed with cclib
        cartesians : list of lists
            List of lists containing the molecular coordinates as floats

        Returns
        -------
        cartesians : list of lists
            New set of cartesian coordinates generated after displacing the original coordinates along the normal modes of the corresponding imaginary frequencies
        """

        shift = []

        # could get rid of atomic units here, if zpe_rat definition is changed
        for mode, _ in enumerate(cclib_data["vibrations"]["frequencies"]):
            # moves along all imaginary freqs (ignoring the TS imag freq, assumed to be the most negative freq)
            if (
                mode == 0
                and cclib_data["metadata"]["ground or transition state"]
                == "transition_state"
            ):
                shift.append(0.0)
            else:
                if cclib_data["vibrations"]["frequencies"][mode] < 0.0:
                    shift.append(float(self.args.amplitude_ifreq))
                else:
                    shift.append(0.0)

            # The starting geometry is displaced along each normal mode according to the random shift
            for atom in range(0, cclib_data["properties"]["number of atoms"]):
                for coord in range(0, 3):
                    cartesians[atom][coord] = (
                        cartesians[atom][coord]
                        + cclib_data["vibrations"]["displacement"][mode][atom][coord]
                        * shift[mode]
                    )

        return cartesians

    def organize_outputs(self, file, termination, errortype, file_terms):
        """
        1. Moves the QM output files to their corresponding folders after the analysis.
        2. Keeps track of the number of calculations with the different types of terminations and error types

        Parameters
        ----------
        file : str
            Output file
        termination : string
            Type of termination of the QM output file (i.e. normal, error, unfinished)
        errortype : string
            Type of error type of the QM output file (i.e. None, not_specified, extra_imag_freq, etc)
        file_terms : dict
            Keeps track of the number of calculations for each termination and error type

        Returns
        -------
        file_terms : dict
            Keeps track of the number of calculations for each termination and error type
        """

        if self.args.resume_qcorr:
            destination_error = self.args.w_dir_main.joinpath(
                f"../../run_{self.args.round_num}/"
            )
            destination_normal = self.args.w_dir_main.joinpath("../../../success/")
        else:
            destination_error = self.args.w_dir_main.joinpath(
                f"failed/run_{self.args.round_num}/"
            )
            destination_normal = self.args.w_dir_main.joinpath("success/")

        if errortype == "none" and termination == "normal":
            destination = destination_normal
            file_terms["finished"] += 1

        elif errortype == "sp_calc" and termination == "normal":
            destination = destination_normal.joinpath("SP_calcs/")
            file_terms["sp_calcs"] += 1

        elif errortype == "extra_imag_freq":
            destination = destination_error.joinpath("extra_imag_freq/")
            file_terms["extra_imag_freq"] += 1

        elif errortype == "ts_no_imag_freq":
            destination = destination_error.joinpath("ts_no_imag_freq/")
            file_terms["ts_no_imag_freq"] += 1

        elif errortype == "spin_contaminated":
            destination = destination_error.joinpath("spin_contaminated/")
            file_terms["spin_contaminated"] += 1

        elif errortype == "duplicate_calc":
            destination = destination_error.joinpath("duplicates/")
            file_terms["duplicate_calc"] += 1

        elif errortype == "atomicbasiserror":
            destination = destination_error.joinpath("error/basis_set_error/")
            file_terms["atom_error"] += 1

        elif errortype == "SCFerror":
            destination = destination_error.joinpath("error/scf_error/")
            file_terms["scf_error"] += 1

        elif errortype == "no_data":
            destination = destination_error.joinpath("error/no_data/")
            file_terms["no_data"] += 1

        elif errortype == "fail_geom_rules":
            destination = destination_error.joinpath("geom_rules_filter/")
            file_terms["geom_rules_qcorr"] += 1

        elif errortype == "isomerization":
            destination = destination_error.joinpath("isomerization/")
            file_terms["isomerized"] += 1

        elif errortype == "freq_no_conv":
            destination = destination_error.joinpath("freq_no_conv/")
            file_terms["freq_no_conv"] += 1

        elif errortype == "linear_mol_wrong":
            destination = destination_error.joinpath("linear_mol_wrong/")
            file_terms["linear_mol_wrong"] += 1

        else:
            destination = destination_error.joinpath("error/not_specified_error/")
            file_terms["not_specified"] += 1

        move_file(destination, self.args.w_dir_main, os.path.basename(file))

        return file_terms, destination

    def write_qcorr_csv(self, file_terms):
        """
        Write information about the QCORR analysis in a csv
        """

        ana_data = pd.DataFrame()
        ana_data.at[0, "Total files"] = len(self.args.files)
        ana_data.at[0, "Normal termination"] = file_terms["finished"]
        ana_data.at[0, "Single-point calcs"] = file_terms["sp_calcs"]
        ana_data.at[0, "Extra imag. freq."] = file_terms["extra_imag_freq"]
        ana_data.at[0, "TS with no imag. freq."] = file_terms["ts_no_imag_freq"]
        ana_data.at[0, "Freq not converged"] = file_terms["freq_no_conv"]
        ana_data.at[0, "Linear mol with wrong n of freqs"] = file_terms[
            "linear_mol_wrong"
        ]
        ana_data.at[0, "SCF error"] = file_terms["scf_error"]
        ana_data.at[0, "No data"] = file_terms["no_data"]
        ana_data.at[0, "Basis set error"] = file_terms["atom_error"]
        ana_data.at[0, "Other errors"] = file_terms["not_specified"]
        if float(self.args.s2_threshold) > 0.0:
            ana_data.at[0, "Spin contamination"] = file_terms["spin_contaminated"]
        ana_data.at[0, "Duplicates"] = file_terms["duplicate_calc"]
        if len(self.args.geom_rules) >= 1:
            ana_data.at[0, "geom_rules filter"] = file_terms["geom_rules_qcorr"]
        if self.args.isom_type is not None:
            ana_data.at[0, "Isomerization"] = file_terms["isomerized"]
        path_as_str = self.args.initial_dir.as_posix()
        csv_qcorr = path_as_str + f"/QCORR-run_{self.args.round_num}-stats.csv"
        ana_data.to_csv(csv_qcorr, index=False)

        return csv_qcorr
